# Copyright (c) 2010-2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager
from io import StringIO
import unittest
from unittest import mock

from swift.common.swob import Request
from storlets.gateway.common.exceptions import FileManagementError
from storlets.swift_middleware.handlers import StorletBaseHandler
from storlets.swift_middleware.handlers.base import get_container_names, \
    SwiftFileManager
from tests.unit import FakeLogger


class TestUtils(unittest.TestCase):
    def setUp(self):
        pass

    def test_get_container_names(self):
        # Use default values
        self.assertEqual(
            {'storlet': 'storlet', 'dependency': 'dependency',
             'log': 'storletlog'},
            get_container_names({}))

        # Use explicit values
        self.assertEqual(
            {'storlet': 'conta', 'dependency': 'contb', 'log': 'contc'},
            get_container_names(
                {'storlet_container': 'conta',
                 'storlet_dependency': 'contb',
                 'storlet_logcontainer': 'contc'}))


class TestSwiftFileManager(unittest.TestCase):
    def setUp(self):
        self.logger = FakeLogger()
        self.manager = SwiftFileManager('a', 'storlet', 'dependency', 'log',
                                        'client.conf', self.logger)

    @contextmanager
    def _mock_internal_client(self, cls):
        with mock.patch('storlets.swift_middleware.handlers.base.'
                        'InternalClient',
                        cls):
            yield

    def test_get_storlet(self):
        name = 'Storlet-1.0.jar'

        class DummyClient(object):
            def __init__(self, *args, **kwargs):
                pass

            def get_object(self, account, container, obj, headers,
                           acceptable_statuses=None):
                return '200', {}, StringIO('test')

        with self._mock_internal_client(DummyClient):
            data_iter, perm = self.manager.get_storlet(name)
            self.assertEqual('test', next(data_iter))
            self.assertIsNone(perm)

        class DummyClient(object):
            def __init__(self, *args, **kwargs):
                pass

            def get_object(self, account, container, obj, headers,
                           acceptable_statuses=None):
                raise Exception('Some error')

        with self._mock_internal_client(DummyClient):
            with self.assertRaises(FileManagementError):
                self.manager.get_storlet(name)

    def test_get_dependency(self):
        name = 'depfile'

        class DummyClient(object):
            def __init__(self, *args, **kwargs):
                pass

            def get_object(self, account, container, obj, headers,
                           acceptable_statuses=None):
                headers = {'X-Object-Meta-Storlet-Dependency-Permissions':
                           '0600'}
                return '200', headers, StringIO('test')

        with self._mock_internal_client(DummyClient):
            data_iter, perm = self.manager.get_dependency(name)
            self.assertEqual('test', next(data_iter))
            self.assertEqual('0600', perm)

        class DummyClient(object):
            def __init__(self, *args, **kwargs):
                pass

            def get_object(self, account, container, obj, headers,
                           acceptable_statuses=None):
                return '200', {}, StringIO('test')

        with self._mock_internal_client(DummyClient):
            data_iter, perm = self.manager.get_dependency(name)
            self.assertEqual('test', next(data_iter))
            self.assertIsNone(perm)

        class DummyClient(object):
            def __init__(self, *args, **kwargs):
                pass

            def get_object(self, account, container, obj, headers,
                           acceptable_statuses=None):
                raise Exception('Some error')

        with self._mock_internal_client(DummyClient):
            with self.assertRaises(FileManagementError):
                self.manager.get_dependency(name)

    def test_put_log(self):
        name = 'logfile'

        class DummyClient(object):
            def __init__(self, *args, **kwargs):
                pass

            def upload_object(self, fobj, account, container, obj,
                              headers=None):
                pass

        with self._mock_internal_client(DummyClient):
            self.manager.put_log(name, mock.MagicMock())

        class DummyClient(object):
            def __init__(self, *args, **kwargs):
                pass

            def upload_object(self, fobj, account, container, obj,
                              headers=None):
                raise Exception('Some error')

        with self._mock_internal_client(DummyClient):
            with self.assertRaises(FileManagementError):
                self.manager.put_log(name, mock.MagicMock())


class TestStorletBaseHandler(unittest.TestCase):

    def test_init_failed_via_base_handler(self):
        def assert_not_implemented(method, path, headers):
            req = Request.blank(
                path, environ={'REQUEST_METHOD': method},
                headers=headers)
            with self.assertRaises(NotImplementedError):
                StorletBaseHandler(
                    req, mock.MagicMock(), mock.MagicMock(),
                    mock.MagicMock(), mock.MagicMock())

        for method in ('PUT', 'GET', 'POST'):
            for path in ('', '/v1', '/v1/a', '/v1/a/c', '/v1/a/c/o'):
                for headers in ({}, {'X-Run-Storlet': 'Storlet-1.0.jar'}):
                    assert_not_implemented(method, path, headers)

    def test_parameters_from_headers(self):
        def mock_request_property():
            """
            This is for skipping extract_vaco which causes NotImplementedError
            if initialize StorletBaseHandler directly.
            """
            def getter(self):
                return self._request

            def setter(self, request):
                self._request = request

            return property(getter, setter, doc='mocked prop')

        headers = {'X-Storlet-Parameter-1': '1:2:3:4',
                   'X-Storlet-parameter-Z': 'A:c'}
        req = Request.blank(
            '/v1/a/c/o', environ={'REQUEST_METHOD': 'GET'},
            headers=headers)

        with mock.patch('storlets.swift_middleware.handlers.base.'
                        'StorletBaseHandler.request', mock_request_property):
            handler = StorletBaseHandler(
                req, mock.MagicMock(), mock.MagicMock(),
                mock.MagicMock(), mock.MagicMock())
            handler._update_storlet_parameters_from_headers()
            self.assertEqual('2:3:4', handler.request.params['1'])
            self.assertEqual('c', handler.request.params['A'])


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2015, 2016 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import StringIO


class FakeStorletFile(object):
    def __init__(self):
        self._call_closed = False

    def close(self):
        self._call_closed = True

    @property
    def closed(self):
        return self._call_closed


class FakeStorletFileIn(FakeStorletFile):
    def __init__(self, input_string, metadata):
        super(FakeStorletFileIn, self).__init__()
        self._input_string = StringIO(input_string)
        self._metadata = metadata
        self._pos = 0

    def read(self, size=-1):
        return self._input_string.read(size)

    def get_metadata(self):
        return self._metadata


class FakeStorletFileOut(FakeStorletFile):
    def __init__(self):
        super(FakeStorletFileOut, self).__init__()
        self._output_string = []
        self._metadata = None

    def write(self, data):
        self._output_string.append(data)

    def set_metadata(self, metadata):
        if self._metadata is not None:
            raise IOError('Sending metadata twice is not allowed')
        self._metadata = {}
        # expect the incoming metadata should be dict
        self._metadata.update(metadata)

    def read(self):
        return ''.join(self._output_string)



# Copyright (c) 2015, 2016 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import errno
import os
import select
import stat
import sys
import time

import eventlet
import json
from contextlib import contextmanager

from storlets.sbus.client import SBusClient
from storlets.sbus.client.exceptions import SBusClientException
from storlets.sbus.datagram import SBusFileDescriptor
from storlets.sbus import file_description as sbus_fd
from storlets.gateway.common.exceptions import StorletRuntimeException, \
    StorletTimeout
from storlets.gateway.common.logger import StorletLogger
from storlets.gateway.common.stob import StorletData, StorletResponse

MAX_METADATA_SIZE = 4096


eventlet.monkey_patch()


class RunTimePaths(object):
    """
    The Storlet Engine need to be access stuff located in many paths:

    1. The various communication channels represented as pipes in the
       filesystem
    2. Directories where to place Storlets
    3. Directories where to place logs

    Communication channels
    ----------------------
    The RunTimeSandbox communicates with the Sandbox via two types of pipes
    1. factory pipe - defined per scope, used for communication with the
       sandbox
       for e.g. start/stop a storlet daemon
    2. Storlet pipe - defined per scope and Storlet, used for communication
       with a storlet daemon, e.g. to call the invoke API

    Each pipe type has two paths:
    1. A path that is inside the sandbox
    2. A path that is outside of the sandbox or at the host side. As such
       this path is prefixed by 'host_'

    Thus, we have the following 4 paths of interest:
    1. sandbox_factory_pipe_path
    2. host_factory_pipe_path
    3. sandbox_storlet_pipe_path
    4. host_storlet_pipe_path

    Our implementation uses the following path structure for the various pipes:
    In the host, all pipes belonging to a given scope are prefixed by
    <pipes_dir>/<scope>, where <pipes_dir> comes from the configuration
    Thus:
    host_factory_pipe_path is of the form <pipes_dir>/<scope>/factory_pipe
    host_storlet_pipe_path is of the form <pipes_dir>/<scope>/<storlet_id>

    In The sandbox side
    sandbox_factory_pipe_path is of the form /mnt/channels/factory_pipe
    sandbox_storlet_pipe_path is of the form  /mnt/channels/<storlet_id>

    Storlets Locations
    ------------------
    The Storlet binaries are accessible from the sandbox using a mounted
    directory.
    This directory is called the storlet directories.
    On the host side it is of the form <storlet_dir>/<scope>/<storlet_name>
    On the sandbox side it is of the form /home/swift/<storlet_name>
    <storlet_dir> comes from the configuration
    <storlet_name> is the prefix of the jar.

    Logs
    ----
    Logs are located in paths of the form:
    <log_dir>/<scope>/<storlet_name>.log
    """

    def __init__(self, scope, conf):
        """
        Construct RunTimePaths instance

        :param scope: scope name to be used as container name
        :param conf: gateway conf
        """
        self.scope = scope
        self.factory_pipe_name = 'factory_pipe'
        self.sandbox_pipe_dir = '/mnt/channels'

        self.sandbox_storlet_base_dir = '/home/swift'
        self.host_root_dir = conf.get('host_root', '/var/lib/storlets')
        self.host_pipe_root_dir = \
            conf.get('pipes_dir',
                     os.path.join(self.host_root_dir, 'pipes', 'scopes'))
        self.host_storlet_root_dir = \
            conf.get('storlets_dir',
                     os.path.join(self.host_root_dir, 'storlets', 'scopes'))
        self.host_log_root_dir = \
            conf.get('log_dir',
                     os.path.join(self.host_root_dir, 'logs', 'scopes'))
        self.host_cache_root_dir = \
            conf.get('cache_dir',
                     os.path.join(self.host_root_dir, 'cache', 'scopes'))

        self.host_storlet_native_lib_dir = '/usr/local/lib/storlets'
        self.sandbox_storlet_native_lib_dir = '/usr/local/lib/storlets'
        self.host_storlet_native_bin_dir = '/usr/local/libexec/storlets'
        self.sandbox_storlet_native_bin_dir = '/usr/local/libexec/storlets'

    @property
    def host_pipe_dir(self):
        return os.path.join(self.host_pipe_root_dir, self.scope)

    def create_host_pipe_dir(self):
        path = self.host_pipe_dir
        if not os.path.exists(path):
            os.makedirs(path)
        # 0777 should be 0700 when we get user namespaces in container
        os.chmod(path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)
        return path

    @property
    def host_factory_pipe(self):
        return os.path.join(self.host_pipe_dir, self.factory_pipe_name)

    @property
    def sandbox_factory_pipe(self):
        return os.path.join(self.sandbox_pipe_dir, self.factory_pipe_name)

    def get_host_storlet_pipe(self, storlet_id):
        return os.path.join(self.host_pipe_dir, storlet_id)

    def get_sbox_storlet_pipe(self, storlet_id):
        return os.path.join(self.sandbox_pipe_dir, storlet_id)

    def get_sbox_storlet_dir(self, storlet_id):
        return os.path.join(self.sandbox_storlet_base_dir, storlet_id)

    @property
    def host_storlet_base_dir(self):
        return os.path.join(self.host_storlet_root_dir, self.scope)

    def get_host_storlet_dir(self, storlet_id):
        return os.path.join(self.host_storlet_base_dir, storlet_id)

    def get_host_slog_path(self, storlet_id):
        return os.path.join(
            self.host_log_root_dir, self.scope, storlet_id,
            'storlet_invoke.log')

    @property
    def host_storlet_cache_dir(self):
        return os.path.join(self.host_cache_root_dir, self.scope, 'storlet')

    @property
    def host_dependency_cache_dir(self):
        return os.path.join(self.host_cache_root_dir, self.scope, 'dependency')


class RunTimeSandbox(object, metaclass=abc.ABCMeta):
    """
    The RunTimeSandbox represents a reusable per scope sandbox.

    The sandbox is reusable in the sense that it can run several storlet
    daemons.

    The following methods are supported:
    ping - pings the sandbox for liveness
    wait - wait for the sandbox to be ready for processing commands
    restart - restart the sandbox
    start_storlet_daemon - start a daemon for a given storlet
    stop_storlet_daemon - stop a daemon of a given storlet
    get_storlet_daemon_status - test if a given storlet daemon is running
    """

    def __init__(self, scope, conf, logger):
        """
        :param scope: scope name to be used as container name
        :param conf: gateway conf
        :param logger: logger instance
        """
        self.paths = RunTimePaths(scope, conf)
        self.scope = scope

        self.sandbox_ping_interval = \
            float(conf.get('sandbox_ping_interval', 0.5))
        self.sandbox_stop_timeout = \
            float(conf.get('stop_linux_container_timeout', 1))
        self.sandbox_wait_timeout = \
            float(conf.get('restart_linux_container_timeout', 10))

        self.container_image_namespace = \
            conf.get('docker_repo', conf.get('container_image_namespace'))
        self.container_image_name_prefix = 'tenant'

        # TODO(add line in conf)
        self.storlet_daemon_thread_pool_size = \
            int(conf.get('storlet_daemon_thread_pool_size', 5))
        self.storlet_daemon_factory_debug_level = \
            conf.get('storlet_daemon_factory_debug_level', 'DEBUG')
        self.storlet_daemon_debug_level = \
            conf.get('storlet_daemon_debug_level', 'DEBUG')

        # TODO(change logger's route if possible)
        self.logger = logger

        self.default_container_image_name = conf.get(
            'default_docker_image_name',
            conf.get('default_container_image_name', 'storlet_engine_image')
        )

        self.max_containers_per_node = \
            int(conf.get('max_containers_per_node', 0))

        self.container_cpu_period = int(conf.get('container_cpu_period', 0))
        self.container_cpu_quota = int(conf.get('container_cpu_quota', 0))
        self.container_mem_limit = conf.get('container_mem_limit', 0)
        # NOTE(tkajinam): memory limit can be a string with unit like 1024m
        try:
            self.container_mem_limit = int(self.container_mem_limit)
        except TypeError:
            pass
        self.container_cpuset_cpus = conf.get('container_cpuset_cpus')
        self.container_cpuset_mems = conf.get('container_cpuset_mems')
        self.container_pids_limit = int(conf.get('container_pids_limit', 0))

    def ping(self):
        """
        Ping to daemon factory process inside container

        :returns: True when the daemon factory is responsive
                  False when the daemon factory is not responsive or it fails
                  to send command to the process
        """
        pipe_path = self.paths.host_factory_pipe
        client = SBusClient(pipe_path)
        try:
            resp = client.ping()
            if not resp.status:
                self.logger.error('Failed to ping to daemon factory: %s' %
                                  resp.message)
            return resp.status
        except SBusClientException:
            return False

    def wait(self):
        """
        Wait while scope's sandbox is starting

        :raises StorletTimeout: the sandbox has not started in
                                sandbox_wait_timeout
        """
        with StorletTimeout(self.sandbox_wait_timeout):
            while not self.ping():
                time.sleep(self.sandbox_ping_interval)

    @abc.abstractmethod
    def _restart(self, container_image_name):
        """
        Restarts the scope's sandbox using the specified container image

        :param container_image_name: name of the container image to start
        :raises StorletRuntimeException: when failed to restart the container
        """
        pass

    def restart(self):
        """
        Restarts the scope's sandbox

        """
        self.paths.create_host_pipe_dir()

        container_image_name = self.scope
        try:
            self._restart(container_image_name)
            self.wait()
        except StorletTimeout:
            raise
        except StorletRuntimeException:
            # We were unable to start a container from the tenant image.
            # Let us try to start a container from default image.
            self.logger.exception("Failed to start a container from "
                                  "tenant image %s" % container_image_name)

            self.logger.info("Trying to start a container from default "
                             "image: %s" % self.default_container_image_name)
            self._restart(self.default_container_image_name)
            self.wait()

    def start_storlet_daemon(
            self, spath, storlet_id, language, language_version=None):
        """
        Start SDaemon process in the scope's sandbox
        """
        pipe_path = self.paths.host_factory_pipe
        client = SBusClient(pipe_path)
        try:
            resp = client.start_daemon(
                language.lower(), spath, storlet_id,
                self.paths.get_sbox_storlet_pipe(storlet_id),
                self.storlet_daemon_debug_level,
                self.storlet_daemon_thread_pool_size,
                language_version)

            if not resp.status:
                self.logger.error('Failed to start storlet daemon: %s' %
                                  resp.message)
                raise StorletRuntimeException('Daemon start failed')
        except SBusClientException:
            raise StorletRuntimeException('Daemon start failed')

    def stop_storlet_daemon(self, storlet_id):
        """
        Stop SDaemon process in the scope's sandbox
        """
        pipe_path = self.paths.host_factory_pipe
        client = SBusClient(pipe_path)
        try:
            resp = client.stop_daemon(storlet_id)
            if not resp.status:
                self.logger.error('Failed to stop storlet daemon: %s' %
                                  resp.message)
                raise StorletRuntimeException('Daemon stop failed')
        except SBusClientException:
            raise StorletRuntimeException('Daemon stop failed')

    def get_storlet_daemon_status(self, storlet_id):
        """
        Get the status of SDaemon process in the scope's sandbox
        """
        pipe_path = self.paths.host_factory_pipe
        client = SBusClient(pipe_path)
        try:
            resp = client.daemon_status(storlet_id)
            if resp.status:
                return 1
            else:
                self.logger.error('Failed to get status about storlet '
                                  'daemon: %s' % resp.message)
                return 0
        except SBusClientException:
            return -1

    def _get_storlet_classpath(self, storlet_main, storlet_id, dependencies):
        """
        Get classpath required to run storlet application

        :param storlet_main: Main class name of the storlet
        :param storlet_id: Name of the storlet file
        :param dependencies: A list of dependency file
        :returns: classpath string
        """
        class_path = os.path.join(
            self.paths.get_sbox_storlet_dir(storlet_main), storlet_id)

        dep_path_list = \
            [os.path.join(self.paths.get_sbox_storlet_dir(storlet_main), dep)
             for dep in dependencies]

        return class_path + ':' + ':'.join(dep_path_list)

    def activate_storlet_daemon(self, sreq, cache_updated=True):
        storlet_daemon_status = \
            self.get_storlet_daemon_status(sreq.storlet_main)
        if (storlet_daemon_status == -1):
            # We failed to send a command to the factory.
            # Best we can do is execute the container.
            self.logger.debug('Failed to check the storlet daemon status. '
                              'Restart its container')
            self.restart()
            storlet_daemon_status = 0

        if (cache_updated is True and storlet_daemon_status == 1):
            # The cache was updated while the daemon is running we need to
            # stop it.
            self.logger.debug('The cache was updated, and the storlet daemon '
                              'is running. Stopping daemon')

            try:
                self.stop_storlet_daemon(sreq.storlet_main)
            except StorletRuntimeException:
                self.logger.warning('Failed to stop the storlet daemon. '
                                    'Restart its container')
                self.restart()
            else:
                self.logger.debug('Deamon stopped')
            storlet_daemon_status = 0

        if (storlet_daemon_status == 0):
            self.logger.debug('Going to start the storlet daemon!')

            # TODO(takashi): This is not needed for python application
            classpath = self._get_storlet_classpath(
                sreq.storlet_main, sreq.storlet_id, sreq.dependencies)

            self.start_storlet_daemon(
                classpath, sreq.storlet_main, sreq.storlet_language,
                sreq.storlet_language_version)
            self.logger.debug('Daemon started')


class StorletInvocationProtocol(object):
    """
    StorletInvocationProtocol class

    This class serves communictaion with a container to run an
    application

    :param srequest: StorletRequest instance
    :param storlet_pipe_path: path string to pipe
    :param storlet_logger_path: path string to log file
    :param timeout: integer of timeout for waiting the resp from container
    :param logger: logger instance
    """
    def __init__(self, srequest, storlet_pipe_path, storlet_logger_path,
                 timeout, logger):
        self.srequest = srequest
        self.storlet_pipe_path = storlet_pipe_path
        self.storlet_logger = StorletLogger(storlet_logger_path)
        self.logger = logger
        self.timeout = timeout

        # local side file descriptors
        self.data_read_fd = None
        self.data_write_fd = None
        self.metadata_read_fd = None
        self.metadata_write_fd = None
        self.task_id = None
        self._input_data_read_fd = None
        self._input_data_write_fd = None

        self.extra_data_sources = []
        for source in self.srequest.extra_data_list:
            if source.has_fd:
                # TODO(kota_): it may be data_fd in the future.
                raise Exception(
                    'extra_source no requires data_fd just data_iter')
            self.extra_data_sources.append(
                {'read_fd': None, 'write_fd': None,
                 'user_metadata': source.user_metadata,
                 'data_iter': source.data_iter})

    @property
    def input_data_read_fd(self):
        """
        File descriptor to read the input body content
        """
        if self.srequest.data.has_fd:
            return self.srequest.data.data_fd
        else:
            return self._input_data_read_fd

    @property
    def remote_fds(self):
        """
        A list of sbus file descriptors passed to remote side
        """
        storlets_metadata = {}
        if self.srequest.has_range:
            storlets_metadata.update(
                {'start': str(self.srequest.start),
                 'end': str(self.srequest.end)})

        fds = [
            SBusFileDescriptor(
                sbus_fd.SBUS_FD_INPUT_OBJECT,
                self.input_data_read_fd,
                storage_metadata=self.srequest.data.user_metadata,
                storlets_metadata=storlets_metadata),
            SBusFileDescriptor(
                sbus_fd.SBUS_FD_OUTPUT_OBJECT,
                self.data_write_fd),
            SBusFileDescriptor(
                sbus_fd.SBUS_FD_OUTPUT_OBJECT_METADATA,
                self.metadata_write_fd),
            SBusFileDescriptor(
                sbus_fd.SBUS_FD_LOGGER,
                self.storlet_logger.getfd())]

        for source in self.extra_data_sources:
            fd = SBusFileDescriptor(
                sbus_fd.SBUS_FD_INPUT_OBJECT,
                source['read_fd'],
                storage_metadata=source['user_metadata'])
            fds.append(fd)

        return fds

    @contextmanager
    def _activate_invocation_descriptors(self):
        """
        Contextmanager about file descriptors used in storlet invocation

        NOTE: This context manager now only closes remote side fds,
              so you should close local side fds
        """
        self._prepare_invocation_descriptors()
        try:
            yield
        finally:
            self._close_remote_side_descriptors()

    def _prepare_invocation_descriptors(self):
        """
        Create all pipse used for Storlet execution
        """
        if not self.srequest.data.has_fd:
            self._input_data_read_fd, self._input_data_write_fd = os.pipe()
        self.data_read_fd, self.data_write_fd = os.pipe()
        self.metadata_read_fd, self.metadata_write_fd = os.pipe()

        for source in self.extra_data_sources:
            source['read_fd'], source['write_fd'] = os.pipe()

    def _safe_close(self, fds):
        """
        Make sure that all of the file descriptors get closed

        :param fds: a list of file descriptors
        """
        for fd in fds:
            try:
                os.close(fd)
            except OSError as err:
                if err.errno != errno.EBADF:
                    raise
                # TODO(kota_): fd might be closed already, so if already
                # closed, OSError will be raised. we need more refactor to
                # keep clean the file descriptors.
                pass

    def _close_remote_side_descriptors(self):
        """
        Close all of the container side descriptors
        """
        fds = [self.data_write_fd, self.metadata_write_fd]
        if self._input_data_read_fd is not None:
            fds.append(self._input_data_read_fd)
        fds.extend([source['read_fd'] for source in self.extra_data_sources])
        for fd in fds:
            os.close(fd)

    def _close_local_side_descriptors(self):
        """
        Close all of the host side descriptors
        """
        fds = [self.data_read_fd, self.metadata_read_fd]
        # NOTE(tkajinam): Local FDs for data input are closed by
        #                 _write_input_data
        self._safe_close(fds)

    def _cancel(self):
        """
        Cancel on-going storlet execution
        """
        client = SBusClient(self.storlet_pipe_path)
        try:
            resp = client.cancel(self.task_id)
            if not resp.status:
                raise StorletRuntimeException('Failed to cancel task')
        except SBusClientException:
            raise StorletRuntimeException('Failed to cancel task')

    def _invoke(self):
        """
        Send an execution command to the remote daemon factory
        """
        with self.storlet_logger.activate(),\
                self._activate_invocation_descriptors():
            self._send_execute_command()

    def _send_execute_command(self):
        """
        Send execute command to the remote daemon factory to invoke storlet
        execution
        """
        client = SBusClient(self.storlet_pipe_path)
        try:
            resp = client.execute(self.srequest.params, self.remote_fds)
            if not resp.status:
                raise StorletRuntimeException("Failed to send execute command")

            if not resp.task_id:
                raise StorletRuntimeException("Missing task id")
            else:
                self.task_id = resp.task_id
        except SBusClientException:
            raise StorletRuntimeException("Failed to send execute command")

    def _wait_for_read_with_timeout(self, fd):
        """
        Wait while the read file descriptor gets ready

        :param fd: File descriptor to read
        :raises StorletTimeout: Exception raised when it times out to cancel
                                the existing task
        :raises StorletRuntimeException: Exception raised when it fails to
                                         cancel the existing task
        """
        try:
            with StorletTimeout(self.timeout):
                r, w, e = select.select([fd], [], [])
        except StorletTimeout:
            exc_type, exc_value, exc_traceback = sys.exc_info()

            # When there is a task already running, we should cancel it.
            if self.task_id:
                try:
                    self._cancel()
                except StorletRuntimeException:
                    self.logger.warning(
                        'Task %s timed out, but failed to get canceled'
                        % self.task_id)
                    pass

            if exc_value is None:
                exc_value = exc_traceback
            if exc_value.__traceback__ is not exc_traceback:
                raise exc_value.with_traceback(exc_traceback)
            raise exc_value

        if fd not in r:
            raise StorletRuntimeException('Read fd is not ready')

    def _read_metadata(self):
        """
        Read metadata in the storlet execution result from fd

        :returns: a dict of metadata
        """
        self._wait_for_read_with_timeout(self.metadata_read_fd)
        flat_json = os.read(self.metadata_read_fd, MAX_METADATA_SIZE)
        os.close(self.metadata_read_fd)
        try:
            return json.loads(flat_json)
        except ValueError:
            self.logger.exception('Failed to load metadata from json')
            raise StorletRuntimeException('Got invalid format about metadata')

    def _wait_for_write_with_timeout(self, fd):
        """
        Wait while the write file descriptor gets ready

        :param fd: File descriptor to write
        :raises StorletTimeout: Exception raised when it times out to cancel
                                the existing task
        :raises StorletRuntimeException: Exception raised when it fails to
                                         cancel the existing task
        """
        with StorletTimeout(self.timeout):
            r, w, e = select.select([], [fd], [])
        if fd not in w:
            raise StorletRuntimeException('Write fd is not ready')

    def communicate(self):
        try:
            self._invoke()

            if not self.srequest.data.has_fd:
                self._wait_for_write_with_timeout(self._input_data_write_fd)

                # We do the writing in a different thread.
                # Otherwise, we can run into the following deadlock
                # 1. middleware writes to Storlet
                # 2. Storlet reads and starts to write metadata and then data
                # 3. middleware continues writing
                # 4. Storlet continues writing and gets stuck as middleware
                #    is busy writing, but still not consuming the reader end
                #    of the Storlet writer.
                eventlet.spawn_n(self._write_input_data,
                                 self._input_data_write_fd,
                                 self.srequest.data.data_iter)

            for source in self.extra_data_sources:
                # NOTE(kota_): not sure right now if using eventlet.spawn_n is
                #              right way. GreenPool is better? I don't get
                #              whole for the dead lock described in above.
                self._wait_for_write_with_timeout(source['write_fd'])
                eventlet.spawn_n(self._write_input_data,
                                 source['write_fd'],
                                 source['data_iter'])

            out_md = self._read_metadata()
            self._wait_for_read_with_timeout(self.data_read_fd)

            data = StorletData(out_md, data_fd=self.data_read_fd,
                               cancel=self._cancel)
            return StorletResponse(data)
        except Exception:
            self._close_local_side_descriptors()
            raise

    @contextmanager
    def _open_writer(self, fd):
        with os.fdopen(fd, 'wb') as writer:
            yield writer

    def _write_input_data(self, fd, data_iter):
        try:
            with self._open_writer(fd) as writer:
                for chunk in data_iter:
                    with StorletTimeout(self.timeout):
                        writer.write(chunk)
        except (OSError, TypeError, ValueError):
            self.logger.exception('fdopen failed')
        except IOError:
            # this will happen at sort of broken pipe while writer.write
            self.logger.exception('IOError with writing fd %s' % fd)
        except StorletTimeout:
            self.logger.exception(
                'Timeout (%s)s with writing fd %s' % (self.timeout, fd))
        except Exception:
            # _write_input_data is designed to run eventlet thread
            # so we should catch an exception and suppress it here
            self.logger.exception('Unexpected error at writing input data')
        finally:
            self._safe_close(fd)



# Copyright (c) 2015, 2016 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import json

from storlets.sbus import command as sbus_cmd
from storlets.sbus import file_description as sbus_fd


class SBusFileDescriptor(object):
    """
    The management class for the file descriptor information
    """

    def __init__(self, fdtype, fileno, storlets_metadata=None,
                 storage_metadata=None):
        self.fdtype = fdtype
        self.fileno = fileno
        self.storlets_metadata = storlets_metadata or {}
        self.storage_metadata = storage_metadata or {}

    @property
    def metadata(self):
        storlets_metadata = copy.deepcopy(self.storlets_metadata)
        storlets_metadata['type'] = self.fdtype
        return {'storlets': storlets_metadata,
                'storage': self.storage_metadata}

    @classmethod
    def from_fileno_and_metadata_dict(cls, fileno, metadict):
        _metadict = copy.deepcopy(metadict)
        storlets_metadata = _metadict['storlets']
        storage_metadata = _metadict['storage']
        fdtype = storlets_metadata.pop('type')
        return cls(fdtype, fileno, storlets_metadata, storage_metadata)


class SBusDatagram(object):
    """
    The manager class for the datagram passed over sbus protocol
    """

    # Each child Datagram should define what fd types are expected with
    # list format
    _required_fdtypes = None

    def __init__(self, command, sfds, params=None, task_id=None):
        """
        Create SBusDatagram instance

        :param command: A string encoding the command to send
        :param sfds: A list of SBusFileDescriptor instances
        :param params: A optional dictionary with parameters for the command
                       execution
        :param task_id: An optional string task id. This is currently used for
                        cancel command
        """
        if type(self) == SBusDatagram:
            raise NotImplementedError(
                'SBusDatagram class should not be initialized as bare')
        self.command = command
        fdtypes = [sfd.fdtype for sfd in sfds]
        self._check_required_fdtypes(fdtypes)
        self.sfds = sfds
        self.params = params
        self.task_id = task_id

    @property
    def num_fds(self):
        return len(self.sfds)

    @property
    def cmd_params(self):
        cmd_params = {'command': self.command}
        if self.params:
            cmd_params['params'] = self.params
        if self.task_id:
            cmd_params['task_id'] = self.task_id
        return cmd_params

    @property
    def serialized_cmd_params(self):
        return json.dumps(self.cmd_params)

    @property
    def fds(self):
        """
        A list of raw file descriptors
        """
        return [sfd.fileno for sfd in self.sfds]

    @property
    def metadata(self):
        return [sfd.metadata for sfd in self.sfds]

    @property
    def serialized_metadata(self):
        return json.dumps(self.metadata)

    @property
    def object_in_metadata(self):
        return [sfd.storage_metadata for sfd in self.sfds
                if sfd.fdtype == sbus_fd.SBUS_FD_INPUT_OBJECT]

    @property
    def object_in_storlet_metadata(self):
        return [fd.storlets_metadata for fd in self.sfds
                if fd.fdtype == sbus_fd.SBUS_FD_INPUT_OBJECT]

    def _find_fds(self, fdtype):
        """
        Get a list of file descriptors for the given fd type

        :param fdtype: file descriptor type
        :returns: a list of file descriptors
        """
        return [sfd.fileno for sfd in self.sfds if sfd.fdtype == fdtype]

    def _find_fd(self, fdtype):
        """
        Get a single file descriptor for the given fd type

        :param fdtype: file descriptor type
        :returns: one file descriptor (integer)
        """
        ret = self._find_fds(fdtype)
        if not ret:
            return None
        else:
            # TODO(takashi): we should raise error if we get multiple fds
            #                for given type. This is to be done when we add
            #                fd validation.
            return ret[0]

    def _check_required_fdtypes(self, given_fdtypes):
        if self._required_fdtypes is None:
            raise NotImplementedError(
                'SBusDatagram class should define _required_fdtypes')
        # the first len(self._required_fdtypes) types should be fit
        # to the required list
        if given_fdtypes[:len(self._required_fdtypes)] != \
                self._required_fdtypes:
            raise ValueError('Fd type mismatch given_fdtypes:%s \
                             required_fdtypes:%s' %
                             (given_fdtypes, self._required_fdtypes))

    def __str__(self):
        return 'num_fds=%s, md=%s, cmd_params=%s' % (
            self.num_fds,
            str(self.serialized_metadata),
            str(self.serialized_cmd_params))


class SBusServiceDatagram(SBusDatagram):
    """
    This class deals with datagram which only has one service out fd.

    This class is used to create datagram for some command type, which
    only needs one service out fd. Currently we can use this class for
    the following commands.
     - SBUS_CMD_HALT
     - SBUS_CMD_START_DAEMON
     - SBUS_CMD_STOP_DAEMON
     - SBUS_CMD_DAEMON_STATUS
     - SBUS_CMD_STOP_DAEMONS
     - SBUS_CMD_PING
     - SBUS_CMD_CANCEL
    """
    _required_fdtypes = [sbus_fd.SBUS_FD_SERVICE_OUT]

    def __init__(self, command, sfds, params=None, task_id=None):
        super(SBusServiceDatagram, self).__init__(
            command, sfds, params, task_id)

    @property
    def service_out_fd(self):
        return self._find_fd(sbus_fd.SBUS_FD_SERVICE_OUT)


class SBusExecuteDatagram(SBusDatagram):
    _required_fdtypes = [sbus_fd.SBUS_FD_SERVICE_OUT,
                         sbus_fd.SBUS_FD_INPUT_OBJECT,
                         sbus_fd.SBUS_FD_OUTPUT_OBJECT,
                         sbus_fd.SBUS_FD_OUTPUT_OBJECT_METADATA,
                         sbus_fd.SBUS_FD_LOGGER]

    def __init__(self, command, sfds, params=None, task_id=None):
        # TODO(kota_): the args command is not used in ExecuteDatagram
        #              but it could be worthful to taransparent init
        #              for other datagram classes.
        # TODO(takashi): When we add extra output sources, we should
        #                consider how we can specify the number of the
        #                extra input/output sources, because currently
        #                this implementation is based on the idea that
        #                we only have extra input sources, which is
        #                added at the end of fd list
        extra_fdtypes = [sfd.fdtype for sfd in
                         sfds[len(self._required_fdtypes):]]

        if [t for t in extra_fdtypes if t != sbus_fd.SBUS_FD_INPUT_OBJECT]:
            raise ValueError(
                'Extra data should be SBUS_FD_INPUT_OBJECT')

        super(SBusExecuteDatagram, self).__init__(
            sbus_cmd.SBUS_CMD_EXECUTE, sfds, params, task_id)

    @property
    def invocation_fds(self):
        return [sfd.fileno for sfd in self.sfds[1:]]

    @property
    def service_out_fd(self):
        return self._find_fd(sbus_fd.SBUS_FD_SERVICE_OUT)

    @property
    def object_out_fds(self):
        return self._find_fds(sbus_fd.SBUS_FD_OUTPUT_OBJECT)

    @property
    def object_metadata_out_fds(self):
        return self._find_fds(sbus_fd.SBUS_FD_OUTPUT_OBJECT_METADATA)

    @property
    def logger_out_fd(self):
        return self._find_fd(sbus_fd.SBUS_FD_LOGGER)

    @property
    def object_in_fds(self):
        return self._find_fds(sbus_fd.SBUS_FD_INPUT_OBJECT)


def build_datagram(command, sfds, params, task_id):
    if command == sbus_cmd.SBUS_CMD_EXECUTE:
        dtg_cls = SBusExecuteDatagram
    else:
        dtg_cls = SBusServiceDatagram

    return dtg_cls(command, sfds, params, task_id)


def build_datagram_from_raw_message(fds, str_md, str_cmd_params):
    """
    Build SBusDatagram from raw message received over sbus

    :param fds: A list of file descriptors (integer) to pass with
                the command
    :param str_md: json serialized metadata dict
    :param str_cmd_params: json serialized command parameters dict
    """
    metadata = json.loads(str_md)
    cmd_params = json.loads(str_cmd_params)
    command = cmd_params.get('command')
    params = cmd_params.get('params')
    task_id = cmd_params.get('task_id')

    if len(fds) != len(metadata):
        raise ValueError('Length mismatch fds: %d != md %d' %
                         (len(fds), len(metadata)))
    sfds = [SBusFileDescriptor.from_fileno_and_metadata_dict(fileno, md)
            for (fileno, md) in zip(fds, metadata)]
    return build_datagram(command, sfds, params, task_id)



# Copyright IBM Corp. 2015, 2015 All Rights Reserved
# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from swiftclient import client


def put_local_file(url, token, container, local_dir, local_file, headers=None):
    """
    Put local file to swift

    :param url: swift endpoint url
    :param token: token string to access to swift
    :param local_dir: directory path where the target file is placed
    :param loca_file: name of the file to be put to swift
    :param headers: headers parameters to be included in request headers
    """
    resp = dict()
    with open(os.path.join(local_dir, local_file), 'rb') as f:
        client.put_object(url, token, container, local_file, f,
                          headers=headers,
                          content_type="application/octet-stream",
                          response_dict=resp)
        status = resp.get('status', 0)
        assert (status // 100 == 2)


def put_storlet_object(url, token, storlet, dependencies, storlet_main_class,
                       language='Java', version=None):
    """
    Put storlet file to swift

    :param url: swift endpoint url
    :param token: token string to access to swift
    :param storlet: storlet file to be registered
    :param dependencies: a list of dependency files
    :param storlet_main_class: name of the storlet main class
    :param language: storlet language. default value is Java
    :param version: storlet language version. default is 3 for python
    """
    headers = {'X-Object-Meta-Storlet-Language': language,
               'X-Object-Meta-Storlet-Interface-Version': '1.0',
               'X-Object-Meta-Storlet-Object-Metadata': 'no',
               'X-Object-Meta-Storlet-Main': storlet_main_class}
    if dependencies:
        headers['X-Object-Meta-Storlet-Dependency'] = dependencies
    if version and language.lower() == 'python':
        headers['X-Object-Meta-Storlet-Language-Version'] = version

    put_local_file(url, token, 'storlet', os.path.dirname(storlet),
                   os.path.basename(storlet), headers)


def put_storlet_executable_dependencies(url, token, deps):
    """
    Put dependency files to swift with 755 permission

    :param url: swift endpoint url:
    :param token: token swring to access to swift
    :param deps: a list of dependency files to be registered
    """
    for dep in deps:
        headers = {'X-Object-Meta-Storlet-Dependency-Version': '1',
                   'X-Object-Meta-Storlet-Dependency-Permissions': '0755'}
        put_local_file(url, token, 'dependency', os.path.dirname(dep),
                       os.path.basename(dep), headers)


def deploy_storlet(url, token, storlet, storlet_main_class, dependencies,
                   language='Java', version=None):
    """
    Deploy storlet file and required dependencies as swift objects

    :param url: swift endpoint url
    :param token: token string to access swift
    :param storlet: storlet file to be registered
    :param dependencies: a list of dependency files to be registered
    :param language: storlet language. default value is Java
    :param version: storlet language version. default is 3 for python
    """
    # No need to create containers every time
    # put_storlet_containers(url, token)
    put_storlet_object(url, token, storlet,
                       ','.join(os.path.basename(x) for x in dependencies),
                       storlet_main_class, language, version)

    put_storlet_executable_dependencies(url, token, dependencies)


def get_auth(conf, user, passwd):
    """
    Get token string to access to swift

    :param conf: a dict of config parameters
    :returns: (swift endpoint url, token string)
    """
    auth_url = conf.auth_uri
    project = conf.project_name
    os_options = {'user_domain_name': conf.domain_name,
                  'project_name': conf.project_name,
                  'region_name': conf.region}
    url, token = client.get_auth(auth_url, project + ':' + user, passwd,
                                 os_options=os_options,
                                 auth_version=conf.auth_version)
    return url, token


def get_admin_auth(conf):
    admin_user = conf.admin_user
    admin_passwd = conf.admin_password
    return get_auth(conf, admin_user, admin_passwd)


def get_member_auth(conf):
    member_user = conf.member_user
    member_passd = conf.member_password
    return get_auth(conf, member_user, member_passd)



# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest
from storlets.gateway.common.exceptions import StorletGatewayLoadError
from storlets.gateway.loader import load_gateway
from storlets.gateway.gateways.stub import StubStorletGateway
from storlets.gateway.gateways.docker import DockerStorletGateway


class TestLoader(unittest.TestCase):
    def setUp(self):
        pass

    def test_load_gateway_entry_point(self):
        # existing entry point
        self.assertEqual(
            StubStorletGateway,
            load_gateway('stub'))

        self.assertEqual(
            DockerStorletGateway,
            load_gateway('docker'))

        # If the given entry point does not exist
        with self.assertRaises(StorletGatewayLoadError):
            load_gateway('foo')

    def test_load_gateway_full_class_path(self):
        # If the given class path exists
        self.assertEqual(
            StubStorletGateway,
            load_gateway('storlets.gateway.gateways.stub.StubStorletGateway'))

        self.assertEqual(
            DockerStorletGateway,
            load_gateway('storlets.gateway.gateways.docker.'
                         'DockerStorletGateway'))

        # If module does not exist
        with self.assertRaises(StorletGatewayLoadError):
            load_gateway('storlets.gateway.gateways.another_stub.'
                         'StubStorletGateway')

        # If class does not exist
        with self.assertRaises(StorletGatewayLoadError):
            load_gateway('storlets.gateway.gateways.stub.'
                         'StorletGatewayAnotherStub')


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2015, 2016 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import docker
import docker.errors
from docker.types import Mount as DockerMount

from storlets.gateway.common.exceptions import StorletRuntimeException
from storlets.gateway.gateways.container.runtime import RunTimeSandbox


class DockerRunTimeSandbox(RunTimeSandbox):

    def _restart(self, container_image_name):
        """
        Restarts the scope's sandbox using the specified container image

        :param container_image_name: name of the container image to start
        :raises StorletRuntimeException: when failed to restart the container
        """
        if self.container_image_namespace:
            container_image_name = '%s/%s' % (self.container_image_namespace,
                                              container_image_name)

        container_name = '%s_%s' % (self.container_image_name_prefix,
                                    self.scope)

        mounts = [
            DockerMount('/dev/log', '/dev/log', type='bind'),
            DockerMount(self.paths.sandbox_pipe_dir,
                        self.paths.host_pipe_dir,
                        type='bind'),
            DockerMount(self.paths.sandbox_storlet_base_dir,
                        self.paths.host_storlet_base_dir,
                        type='bind'),
            DockerMount(self.paths.sandbox_storlet_native_lib_dir,
                        self.paths.host_storlet_native_lib_dir,
                        type='bind', read_only=True),
            DockerMount(self.paths.sandbox_storlet_native_bin_dir,
                        self.paths.host_storlet_native_bin_dir,
                        type='bind', read_only=True)
            ]

        try:
            client = docker.from_env()
            # Stop the existing storlet container
            try:
                scontainer = client.containers.get(container_name)
            except docker.errors.NotFound:
                # The container is not yet created
                pass
            else:
                scontainer.stop(timeout=self.sandbox_stop_timeout)

            # Check whether a new container can be started
            if self.max_containers_per_node > 0:
                all_scontainers = client.containers.list(
                    filters={'label': 'managed_by=storlets'})
                if len(all_scontainers) >= self.max_containers_per_node:
                    raise StorletRuntimeException(
                        "Cannot start a container because of limit")

            # Start the new one
            client.containers.run(
                container_image_name, detach=True,
                command=[
                    self.paths.sandbox_factory_pipe,
                    self.storlet_daemon_factory_debug_level,
                    container_name
                ],
                entrypoint=[
                    os.path.join(
                        self.paths.sandbox_storlet_native_bin_dir,
                        'storlets-daemon-factory'
                    )
                ],
                environment={
                    'PYTHONPATH': os.path.join(
                        self.paths.sandbox_storlet_native_lib_dir, 'python'
                    )
                },
                name=container_name, network_disabled=True,
                mounts=mounts, user=os.getuid(),
                auto_remove=True, stop_signal='SIGHUP',
                cpu_period=self.container_cpu_period,
                cpu_quota=self.container_cpu_quota,
                mem_limit=self.container_mem_limit,
                cpuset_cpus=self.container_cpuset_cpus,
                cpuset_mems=self.container_cpuset_mems,
                pids_limit=self.container_pids_limit,
                labels={'managed_by': 'storlets'})
        except docker.errors.ImageNotFound:
            msg = "Image %s is not found" % container_image_name
            raise StorletRuntimeException(msg)
        except docker.errors.APIError:
            self.logger.exception("Failed to manage docker containers")
            raise StorletRuntimeException("Docker runtime error")



# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from swiftclient import client as c
from tests.functional.java import StorletJavaFunctionalTest
import unittest


class TestCsvStorlet(StorletJavaFunctionalTest):
    def setUp(self):
        self.storlet_log = 'csvstorlet-1.0.log'
        self.additional_headers = {}
        main_class = 'org.openstack.storlet.csv.CSVStorlet'
        super(TestCsvStorlet, self).setUp('CsvStorlet',
                                          'csvstorlet-1.0.jar',
                                          main_class,
                                          'meter-1MB.csv')

    def invoke_storlet(self, start, end,
                       first_partition, max_record_line,
                       columns_selection,
                       where_clause):
        headers = {'X-Run-Storlet': self.storlet_name}
        headers.update(self.additional_headers)
        headers['X-Storlet-Range'] = 'bytes=%d-%d' % (start,
                                                      end + max_record_line)
        headers['X-Storlet-Parameter-1'] = '%s:%s' % ('start', start)
        headers['X-Storlet-Parameter-2'] = '%s:%s' % ('end', end)
        headers['X-Storlet-Parameter-3'] = '%s:%s' % ('max_record_line',
                                                      max_record_line)
        headers['X-Storlet-Parameter-4'] = '%s:%s' % ('first_partition',
                                                      first_partition)
        headers['X-Storlet-Parameter-5'] = '%s:%s' % ('selected_columns',
                                                      columns_selection)
        headers['X-Storlet-Parameter-6'] = '%s:%s' % ('where_clause',
                                                      where_clause)
        _, content = c.get_object(
            self.url, self.token, self.container, self.storlet_file,
            response_dict=dict(), headers=headers)
        content = content.decode('utf-8')
        return content

    def _test_filter(self):
        content = self.invoke_storlet(120, 1024,
                                      'false', 512,
                                      '4,6', '')
        for line in content.split('\n'):
            if line:
                self.assertEqual(2, len(line.split(',')))

    def _test_prune(self):
        content = self.invoke_storlet(120, 1024,
                                      'false', 512,
                                      '',
                                      'EqualTo(6,ESP)')
        for line in content.split('\n'):
            if line:
                val = line.split(',')[6]
                self.assertEqual('ESP', val)

    def _test_prune_filter(self):
        content = self.invoke_storlet(120, 1024,
                                      'false', 512,
                                      '4,6',
                                      'EqualTo(6,ESP)')
        for line in content.split('\n'):
            if line:
                val = line.split(',')[1]
                self.assertEqual('ESP', val)

    def _prune_filter_scan_with_count(self, start, stop,
                                      first, max_len,
                                      columns, where):
        content = self.invoke_storlet(start, stop,
                                      first, max_len,
                                      columns, where)
        count = 0
        for line in content.split('\n'):
            if line:
                self.assertEqual('FRA', line.split(',')[1])
                count = count + 1
        return count

    def test_complete_file(self):
        max_len = 512
        c1 = self._prune_filter_scan_with_count(58, 349557,
                                                'true', max_len,
                                                '4,6', 'EqualTo(6,FRA)')
        c2 = self._prune_filter_scan_with_count(349558, 699057,
                                                'false', max_len,
                                                '4,6', 'EqualTo(6,FRA)')
        c3 = self._prune_filter_scan_with_count(699058, 1048558,
                                                'false', max_len,
                                                '4,6', 'EqualTo(6,FRA)')
        self.assertEqual(1070, c1 + c2 + c3)


class TestCsvStorletOnProxy(TestCsvStorlet):
    def setUp(self):
        super(TestCsvStorletOnProxy, self).setUp()
        self.additional_headers = {'X-Storlet-Run-On-Proxy': ''}


if __name__ == '__main__':
    unittest.main()



# Copyright IBM Corp. 2015, 2015 All Rights Reserved
# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from swiftclient import client as c
from tests.functional.java import StorletJavaFunctionalTest
import unittest


class TestMetadataStorlet(StorletJavaFunctionalTest):
    def setUp(self):
        self.storlet_log = 'testmetadatastorlet-1.0.log'
        headers = {'X-Object-Meta-key1': '1',
                   'X-Object-Meta-key2': '2',
                   'X-Object-Meta-key3': '3',
                   'X-Object-Meta-key4': '4',
                   'X-Object-Meta-key5': '5',
                   'X-Object-Meta-key6': '6',
                   'X-Object-Meta-key7': '7',
                   'X-Object-Meta-key8': '8',
                   'X-Object-Meta-key9': '9',
                   'X-Object-Meta-key10': '10'}
        self.additional_headers = {}
        main_class = ('org.openstack.storlet.testmetadatastorlet'
                      '.MetadataStorlet')
        super(TestMetadataStorlet, self).setUp('TestMetadataStorlet',
                                               'testmetadatastorlet-1.0.jar',
                                               main_class,
                                               'source.txt',
                                               headers=headers)

    def test_metadata_get(self, params=None, global_params=None):
        if params is not None:
            querystring = ''
            for key in params:
                querystring += '%s=%s,' % (key, params[key])
            querystring = querystring[:-1]
        else:
            querystring = None

        headers = {'X-Run-Storlet': self.storlet_name}
        headers.update(self.additional_headers)
        original_headers, original_content = \
            c.get_object(self.url, self.token,
                         self.container, self.storlet_file,
                         response_dict=dict(), headers=headers)
        self.assertEqual('1', original_headers['X-Object-Meta-key1'.lower()])
        self.assertEqual('2', original_headers['X-Object-Meta-key2'.lower()])
        self.assertEqual('3', original_headers['X-Object-Meta-key3'.lower()])
        self.assertEqual('4', original_headers['X-Object-Meta-key4'.lower()])
        self.assertEqual('5', original_headers['X-Object-Meta-key5'.lower()])
        self.assertEqual('6', original_headers['X-Object-Meta-key6'.lower()])
        self.assertEqual('7', original_headers['X-Object-Meta-key7'.lower()])
        self.assertEqual('8', original_headers['X-Object-Meta-key8'.lower()])
        self.assertEqual('9', original_headers['X-Object-Meta-key9'.lower()])
        self.assertEqual('10', original_headers['X-Object-Meta-key10'.lower()])
        omv = original_headers['X-Object-Meta-override_key'.lower()]
        self.assertEqual('new_value', omv)


class TestMetadataStorletOnProxy(TestMetadataStorlet):
    def setUp(self):
        super(TestMetadataStorletOnProxy, self).setUp()
        self.additional_headers = {'X-Storlet-Run-On-Proxy': ''}


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random


class MultiInputMIMEStorlet(object):
    def __init__(self, logger):
        self.logger = logger

    def __call__(self, in_files, out_files, params):
        """
        The function called for storlet invocation

        :param in_files: a list of StorletInputFile
        :param out_files: a list of StorletOutputFile
        :param params: a dict of request parameters
        """

        metadata = {}
        for input_file in in_files:
            metadata.update(input_file.get_metadata())
        mime_boundary = "%.64x" % random.randint(0, 16 ** 64)
        metadata['Content-Type'] = \
            'multipart/mixed; boundary=%s' % mime_boundary
        out_files[0].set_metadata(metadata)

        self.logger.debug('Start to return object data')
        while in_files:
            input_file = in_files.pop(0)
            while True:
                buf = input_file.read(16)
                if not buf:
                    break
                self.logger.debug('Received %d bytes' % len(buf))
                out_files[0].write(buf)
            input_file.close()
            if in_files:
                # in_files still have items
                out_files[0].write(
                    ('\n--%s\n' % mime_boundary).encode('utf-8'))
            else:
                # this is the end of input_files so the boundary should end
                # the content
                out_files[0].write(
                    ('\n--%s--' % mime_boundary).encode('utf-8'))

        self.logger.debug('Complete')
        out_files[0].close()



# Copyright IBM Corp. 2015, 2015 All Rights Reserved
# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from storlets.gateway.common.stob import StorletRequest, \
    StorletResponse, StorletData
from storlets.gateway.gateways.base import StorletGatewayBase


class StubStorletGateway(StorletGatewayBase):

    request_class = StorletRequest

    def __init__(self, conf, logger, scope):
        super(StubStorletGateway, self).__init__(conf, logger, scope)
        self.logger = logger
        self.conf = conf
        self.scope = scope

    @classmethod
    def validate_storlet_registration(cls, params, obj):
        pass

    @classmethod
    def validate_dependency_registration(cls, params, obj):
        pass

    def indentity_invocation(self, sreq):
        self.logger.debug("Identity invocation is called")
        data = StorletData(sreq.data.user_metadata, sreq.data.data_iter)
        return StorletResponse(data)

    def invocation_flow(self, sreq, extra_resources=None):
        return self.indentity_invocation(sreq)



# Copyright (c) 2015-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import partial
import json
import os
import signal

from storlets.sbus import SBus
import storlets.sbus.command as sbus_cmd

EXIT_SUCCESS = 0
EXIT_FAILURE = 1

LOOP_TIMEOUT = 0.5
LISTEN_TIMEOUT = 300


class CommandResponse(Exception):
    """
    The result of command execution
    """

    def __init__(self, status, message, iterable=True, task_id=None):
        """
        Construct CommandResponse instance

        :param status: task status
        :param message: message to be returned and logged
        :param iterable: whether we can keep SDaemon process running
        :param task_id: ID assigned to the requested task
        """
        self.status = status
        self.message = message

        # NOTE(takashi): iterable controls whether the server main loop should
        #                exit or not as a result of processing the command
        self.iterable = iterable

        self.task_id = task_id

    @property
    def report_message(self):
        rsp = {'status': self.status, 'message': self.message}
        if self.task_id:
            rsp['task_id'] = self.task_id
        return json.dumps(rsp)


CommandSuccess = partial(CommandResponse, True)
CommandFailure = partial(CommandResponse, False)


def command_handler(func):
    """
    Decorator for handler methods called according to given sbus command
    """
    func.is_command_handler = True
    return func


class SBusServer(object):
    """
    Basic implementation for server processes listening on sbus protocol

    NOTE(takash): we should implement some termination functions like
                  _terminate and halt, otherwise they raise
                  NotImplementedError.
    """
    def __init__(self, sbus_path, logger):
        self.sbus_path = sbus_path
        self.logger = logger
        self.running = True
        self.listen_timeout = LISTEN_TIMEOUT
        self.loop_timeout = LOOP_TIMEOUT

    def get_handler(self, command):
        """
        Decide handler function corresponding to the received command

        :param command: command
        :returns: handler function
        """
        # All commands should be defined in storlets.sbus.command
        if not hasattr(sbus_cmd, command):
            raise ValueError('Unknown command %s' % command)

        func_name = command[len(sbus_cmd.SBUS_CMD_PREFIX):].lower()
        try:
            handler = getattr(self, func_name)
            getattr(handler, 'is_command_handler')
        except AttributeError:
            raise ValueError('Command %s is not allowed for this server' %
                             command)
        return handler

    def dispatch_command(self, dtg):
        """
        Parse datagram. React on the request.

        :param dtg: Datagram received from client

        :returns: True if the server can continue its main loop
                  False if the server should terminate its main loop
        """
        command = dtg.command
        self.logger.debug("Received command %s" % command)

        try:
            handler = self.get_handler(command)
            resp = handler(dtg)
        except ValueError as err:
            self.logger.exception('Failed to handle request')
            resp = CommandFailure(str(err))
        except CommandResponse as err:
            resp = err
        except Exception:
            self.logger.exception('Failed to handle request')
            resp = CommandFailure('Internal error')

        self.logger.info('Command:%s Response:%s' %
                         (command, resp.report_message))

        outfd = dtg.service_out_fd
        with os.fdopen(outfd, 'wb') as outfile:
            self._respond(outfile, resp)

        return resp.iterable

    def _respond(self, outfile, resp):
        """
        Send result description message back to gateway

        :param outfile : Output channel to send the message to
        :param resp: CommandResponse instance
        """
        try:
            outfile.write(resp.report_message.encode('utf-8'))
        except IOError:
            self.logger.exception('Unable to return response to client')

    @command_handler
    def ping(self, dtg):
        self.logger.debug('Received ping. Responding')
        return CommandSuccess('OK')

    @command_handler
    def halt(self, dtg):
        raise NotImplementedError()

    def _terminate(self):
        raise NotImplementedError()

    def _force_terminate(self):
        raise NotImplementedError()

    def _graceful_shutdown(self, *args):
        self.logger.info("received SIGHUP. Shutting down gracefully")
        self.running = False

    def _force_shutdown(self, *args):
        self.logger.info("received SIGTERM. Shutting down now")
        try:
            self._force_terminate()
        except Exception:
            self.logger.exception(
                "Failed to force_terminate. The process stops anyway")
        finally:
            os._exit(1)

    def main_loop(self):
        """
        Main loop to run storlet application

        :returns: EXIT_SUCCESS when the loop exists normally
                  EXIT_FAILURE when an error occurs in main loop
        """
        sbus = SBus()
        fd = sbus.create(self.sbus_path)
        if fd < 0:
            self.logger.error("Failed to create SBus. exiting.")
            return EXIT_FAILURE

        signal.signal(signal.SIGHUP, self._graceful_shutdown)
        signal.signal(signal.SIGTERM, self._force_shutdown)

        loop_cnt = 0
        status = EXIT_SUCCESS

        while self.running:
            rc = sbus.listen(fd, self.loop_timeout)

            if rc < 0:
                self.logger.error("Failed to wait on SBus. exiting.")
                status = EXIT_FAILURE
                break
            elif rc == 0:
                loop_cnt += 1
                if loop_cnt * self.loop_timeout >= self.listen_timeout:
                    self.logger.debug("Timed out while listening. exiting.")
                    break
                continue

            # Reset loop_cnt here as a request comes via sbus
            loop_cnt = 0

            dtg = sbus.receive(fd)
            if dtg is None:
                self.logger.error("Failed to receive message. exiting")
                status = EXIT_FAILURE
                break

            if not self.dispatch_command(dtg):
                break

        self.logger.debug('Leaving main loop')
        self._terminate()

        return status



# Copyright (c) 2016 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
from storlets.sbus import SBus
from storlets.sbus import command as sbus_cmd
from storlets.sbus.datagram import SBusFileDescriptor, build_datagram
from storlets.sbus.file_description import SBUS_FD_SERVICE_OUT
from storlets.sbus.client.exceptions import SBusClientIOError, \
    SBusClientMalformedResponse, SBusClientSendError


class SBusResponse(object):
    def __init__(self, status, message, task_id=None):
        """
        Construct SBusResponse class

        :param status: Whether the server succeed to process the given request
        :param message: Messages to describe the process result
        :param task_id: Id assigned to each tasks
        """
        self.status = status
        self.message = message
        self.task_id = task_id


class SBusClient(object):
    def __init__(self, socket_path, chunk_size=16):
        self.socket_path = socket_path
        self.chunk_size = chunk_size

    def _parse_response(self, str_response):
        """
        Parse response string received from container side

        :param str_response: response string
        :returns: SBusResponse instance
        """
        try:
            resp = json.loads(str_response)
            status = resp['status']
            message = resp['message']
        except (ValueError, KeyError):
            raise SBusClientMalformedResponse('Got malformed response')

        # NOTE(takashi): task_id is currently used only in EXECUTE command, so
        #                we don't fail here even if the given response doesn't
        #                have task_id.
        task_id = resp.get('task_id')

        return SBusResponse(status, message, task_id)

    def _request(self, command, params=None, task_id=None, extra_fds=None):
        read_fd, write_fd = os.pipe()
        try:
            try:
                sfds = \
                    [SBusFileDescriptor(SBUS_FD_SERVICE_OUT, write_fd)] + \
                    (extra_fds or [])
                datagram = build_datagram(command, sfds, params, task_id)
                rc = SBus.send(self.socket_path, datagram)
                if rc < 0:
                    raise SBusClientSendError(
                        'Faild to send command(%s) to socket %s' %
                        (datagram.command, self.socket_path))
            finally:
                # We already sent the write fd to remote, so should close it
                # in local side before reading response
                os.close(write_fd)

            reply = b''
            while True:
                try:
                    buf = os.read(read_fd, self.chunk_size)
                except IOError:
                    raise SBusClientIOError(
                        'Failed to read data from read pipe')
                if not buf:
                    break
                reply = reply + buf
        finally:
            os.close(read_fd)

        return self._parse_response(reply.decode('utf-8'))

    def execute(self, invocation_params, invocation_fds):
        return self._request(sbus_cmd.SBUS_CMD_EXECUTE,
                             params=invocation_params,
                             extra_fds=invocation_fds)

    def ping(self):
        return self._request(sbus_cmd.SBUS_CMD_PING)

    def start_daemon(self, language, storlet_path, storlet_id,
                     uds_path, log_level, pool_size,
                     language_version):
        params = {'daemon_language': language, 'storlet_path': storlet_path,
                  'storlet_name': storlet_id, 'uds_path': uds_path,
                  'log_level': log_level, 'pool_size': pool_size}
        if language_version:
            params['daemon_language_version'] = language_version

        return self._request(sbus_cmd.SBUS_CMD_START_DAEMON, params)

    def stop_daemon(self, storlet_name):
        return self._request(sbus_cmd.SBUS_CMD_STOP_DAEMON,
                             {'storlet_name': storlet_name})

    def stop_daemons(self):
        return self._request(sbus_cmd.SBUS_CMD_STOP_DAEMONS)

    def halt(self):
        return self._request(sbus_cmd.SBUS_CMD_HALT)

    def daemon_status(self, storlet_name):
        return self._request(sbus_cmd.SBUS_CMD_DAEMON_STATUS,
                             {'storlet_name': storlet_name})

    def cancel(self, task_id):
        return self._request(sbus_cmd.SBUS_CMD_CANCEL, task_id=task_id)



# Copyright (c) 2010-2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest import mock

from storlets.gateway.gateways.stub import StubStorletGateway
from storlets.swift_middleware import storlet_handler

from tests.unit import FakeLogger
from tests.unit.swift_middleware import FakeApp


def create_handler_config(exec_server):
    return {'execution_server': exec_server,
            'gateway_module': StubStorletGateway}


class BaseTestStorletMiddleware(unittest.TestCase):
    def setUp(self, exec_server='proxy'):
        self.exec_server = exec_server
        self.conf = create_handler_config(exec_server)
        self.logger = FakeLogger()
        self.base_app = FakeApp()

    def tearDown(self):
        pass

    def get_app(self, app, global_conf, **local_conf):
        with mock.patch('storlets.swift_middleware.storlet_handler.'
                        'get_logger') as get_fake_logger:
            get_fake_logger.return_value = self.logger
            factory = storlet_handler.filter_factory(global_conf, **local_conf)
            return factory(app)

    def get_response(self, req):
        app = self.get_app(self.base_app, self.conf)
        return req.get_response(app)



# Copyright (c) 2010-2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from swift.common.middleware.copy import \
    _check_copy_from_header as check_copy_from_header, \
    _check_destination_header as check_destination_header, \
    _copy_headers as copy_headers
from swift.common.swob import HTTPBadRequest, HTTPUnauthorized, \
    HTTPMethodNotAllowed, HTTPPreconditionFailed, HTTPForbidden
from swift.common.utils import config_true_value, public, FileLikeIter, \
    list_from_csv, split_path
from swift.common.middleware.acl import clean_acl
from swift.common.wsgi import make_subrequest
from swift.proxy.controllers.base import get_account_info
from urllib.parse import quote

from storlets.gateway.common.stob import StorletData
from storlets.swift_middleware.handlers.base import StorletBaseHandler, \
    NotStorletRequest, NotStorletExecution


CONDITIONAL_KEYS = ['IF_MATCH', 'IF_NONE_MATCH', 'IF_MODIFIED_SINCE',
                    'IF_UNMODIFIED_SINCE']

REFERER_PREFIX = 'storlets'


class StorletProxyHandler(StorletBaseHandler):
    def __init__(self, request, conf, gateway_conf, app, logger):
        super(StorletProxyHandler, self).__init__(
            request, conf, gateway_conf, app, logger)
        self.max_extra_resources = int(conf.get('max_extra_resources', -1))
        self.storlet_containers = [self.storlet_container,
                                   self.storlet_dependency]
        self.agent = 'ST'

        # A very initial hook for blocking requests
        self._should_block(request)

        if not self.is_storlet_request:
            # This is not storlet-related request, so pass it
            raise NotStorletRequest()

        # In proxy server, storlet handler validate if storlet enabled
        # at the account, anyway
        account_meta = get_account_info(self.request.environ,
                                        self.app)['meta']
        storlets_enabled = account_meta.get('storlet-enabled', 'False')
        if not config_true_value(storlets_enabled):
            msg = 'Account disabled for storlets'
            self.logger.debug(msg)
            raise HTTPBadRequest(msg.encode('utf8'), request=self.request)

        if self.is_storlet_acl_update:
            self.acl_string = self._validate_acl_update(self.request)
        elif self.is_storlet_object_update:
            # TODO(takashi): We have to validate metadata in COPY case
            self._validate_registration(self.request)
            raise NotStorletExecution()
        elif self.is_storlet_execution:
            self._setup_gateway()
            self.extra_resources = self._parse_extra_resources()
        else:
            raise NotStorletExecution()

    def _should_block(self, request):
        # Currently, we have only one reason to block
        # requests at such an early stage of the processing:
        # we block requests with referer that have the internal prefix
        # of:
        if request.referer and REFERER_PREFIX in request.referer:
            msg = 'Referrer containing %s is not allowed' % REFERER_PREFIX
            self.logger.debug(msg)
            raise HTTPForbidden(msg.encode('utf8'), request=self.request)

    def _parse_vaco(self):
        return self.request.split_path(3, 4, rest_with_last=True)

    def is_proxy_runnable(self, resp=None):
        """
        Check if the storlet should be executed at proxy server

        :param resp: swob.Response instance
        :return: Whether we should execute the storlet at proxy
        """
        checks = [
            self.execute_on_proxy,
            self.execute_range_on_proxy
        ]
        if resp:
            checks.extend([
                self.is_slo_response(resp),
                self.is_symlink_response(resp)
            ])

        runnable = any(checks)
        return runnable

    @property
    def is_storlet_request(self):
        return (self.is_storlet_execution or self.is_storlet_object_update
                or self.is_storlet_acl_update)

    @property
    def is_storlet_object_update(self):
        return (self.container in self.storlet_containers and self.obj
                and self.request.method in ['PUT', 'POST'])

    @property
    def is_storlet_acl_update(self):
        return (self.request.method == 'POST' and not self.obj and
                'X-Storlet-Container-Read' in self.request.headers)

    @property
    def is_put_copy_request(self):
        return 'X-Copy-From' in self.request.headers

    def is_symlink_response(self, resp):
        """
        Determines whether the response is a symlink one

        :param resp: swob.Response instance
        :return: Whenther the response is a slo one
        """
        self.logger.debug(
            'Verify if {0} is a symlink'.format(self.path))

        symlink_header = 'X-Symlink-Target'
        is_symlink = symlink_header in resp.headers
        if is_symlink:
            self.logger.debug(
                '{0} is indeed a symlink'.format(self.path))
        else:
            self.logger.debug(
                '{0} is NOT a symlink'.format(self.path))
        return is_symlink

    def _parse_storlet_params(self, headers):
        """
        Parse storlet parameters from storlet/dependency object metadata

        :returns: dict of storlet parameters
        """
        params = dict()
        for key in headers:
            if key.startswith('X-Object-Meta-Storlet-'):
                params[key[len('X-Object-Meta-Storlet-'):]] = headers[key]
        return params

    def _validate_registration(self, req):
        """
        Validate parameters about storlet/dependency object when registering

        :params req: swob.Request instance
        :raises ValueError: If some parameters are wrong
        """
        params = self._parse_storlet_params(req.headers)
        try:
            if self.container == self.storlet_container:
                self.logger.debug('updating object in storlet container. '
                                  'Sanity check')
                self.gateway_class.validate_storlet_registration(
                    params, self.obj)
            else:
                self.logger.debug('updating object in storlet dependency. '
                                  'Sanity check')
                self.gateway_class.validate_dependency_registration(
                    params, self.obj)
        except ValueError as e:
            self.logger.exception('Bad parameter')
            raise HTTPBadRequest(e.args[0].encode('utf8'))

    def _build_acl_string(self, user, storlet):
        acl_string = '%s.%s_%s' % (REFERER_PREFIX, user, storlet)
        return acl_string

    def _validate_acl_update(self, req):
        """
        Validate the request has the necessary headers for a
        storlet ACL update

        :params req: swob.Request instance
        :return: the resulting acl string that should be added
        :raises HTTPBadRequest: If a header is missing or mulformed
        """
        # Make sure we are not meddling with the storlet containers
        if self.container in self.storlet_containers:
            msg = b'storlet ACL update cannot be a storlet container'
            raise HTTPBadRequest(msg)

        # Make sure the expected headers are supplied
        user_name = req.headers.get("X-Storlet-Container-Read", None)
        storlet_name = req.headers.get("X-Storlet-Name", None)
        if not user_name or not storlet_name:
            msg = b'storlet ACL update request is missing a mandatory header'
            raise HTTPBadRequest(msg)

        # Make sure the resulting acl is valid
        acl_string = '.r:%s' % self._build_acl_string(user_name, storlet_name)
        try:
            clean_acl('X-Container-Read', acl_string)
        except ValueError as err:
            msg = ('storlet ACL update request has invalid values %s'
                   % str(err))
            raise HTTPBadRequest(msg.encode('utf8'))

        # Make sure the resulting acl permits a single entity
        if ',' in acl_string:
            msg = 'storlet ACL update request has ' \
                'mulformed storlet or user name'
            raise HTTPBadRequest(msg.encode('utf8'))

        # The request is valid. Keep the ACL string
        return acl_string

    def verify_access_to_storlet(self):
        """
        Verify access to the storlet object

        :return: storlet parameters
        :raises HTTPUnauthorized: If it fails to verify access
        """
        sobj = self.request.headers.get('X-Run-Storlet')
        spath = '/'.join(['', self.api_version, self.account,
                          self.storlet_container, sobj])
        self.logger.debug('Verify access to %s' % spath)

        new_env = dict(self.request.environ)
        if 'HTTP_TRANSFER_ENCODING' in new_env:
            del new_env['HTTP_TRANSFER_ENCODING']

        for key in CONDITIONAL_KEYS:
            env_key = 'HTTP_' + key
            if env_key in new_env:
                del new_env[env_key]

        auth_token = self.request.headers.get('X-Auth-Token')
        storlet_req = make_subrequest(
            new_env, 'HEAD', spath,
            headers={'X-Auth-Token': auth_token},
            swift_source=self.agent)

        resp = storlet_req.get_response(self.app)
        if not resp.is_success:
            msg = 'Failed to verify access to the storlet. ' \
                  'Either the storlet does not exist or ' \
                  'you are not authorized to run the ' \
                  'storlet.'
            raise HTTPUnauthorized(msg.encode('utf8'),
                                   request=self.request)

        params = self._parse_storlet_params(resp.headers)
        for key in ['Content-Length', 'X-Timestamp']:
            params[key] = resp.headers[key]
        return params

    def handle_request(self):
        if hasattr(self, self.request.method):
            try:
                handler = getattr(self, self.request.method)
                getattr(handler, 'publicly_accessible')
            except AttributeError:
                # TODO(kota_): add allowed_method list to Allow header
                return HTTPMethodNotAllowed(request=self.request)
            return handler()
        else:
            raise HTTPMethodNotAllowed(request=self.request)

    def _call_gateway(self, resp):
        sreq = self._build_storlet_request(self.request, resp.headers,
                                           resp.app_iter)
        return self.gateway.invocation_flow(sreq)

    def augment_storlet_request(self, params):
        """
        Add to request the storlet parameters to be used in case the request
        is forwarded to the data node (GET case)

        :param params: parameters to be augmented to request
        """
        for key, val in params.items():
            self.request.headers['X-Storlet-' + key] = val

    def _parse_extra_resources(self):
        if 'X-Storlet-Extra-Resources' not in self.request.headers:
            return []

        try:
            resources = list_from_csv(
                self.request.headers['X-Storlet-Extra-Resources'])

            if len(resources) == 0:
                return []

            if self.max_extra_resources > 0:
                if len(resources) > self.max_extra_resources:
                    msg = (
                        'Too many extra resources. %d was requested but only '
                        '%s is allowed,' %
                        (len(resources), self.max_extra_resources)
                    )
                    raise HTTPBadRequest(msg.encode('utf-8'),
                                         request=self.request)
            elif self.max_extra_resources == 0:
                msg = 'Usage of extra resources is not allowed'
                raise HTTPForbidden(msg.encode('utf-8'), request=self.request)

            res_list = []
            # resourece should be /container/object
            for resource in resources:
                # sanity check, if it's invalid path ValueError
                # will be raisen
                swift_path = ['', self.api_version, self.account]
                swift_path.extend(split_path(resource, 2, 2, True))
                res_list.append('/'.join(swift_path))
            return res_list
        except ValueError:
            msg = ('X-Storlet-Extra-Resource must be a csv with'
                   '/container/object format')
            raise HTTPBadRequest(msg.encode('utf-8'), request=self.request)

    def _build_storlet_request(self, req, sheaders, sbody_iter):
        sreq = super(StorletProxyHandler, self)._build_storlet_request(
            req, sheaders, sbody_iter)
        for resource in self.extra_resources:
            # TODO(kota_): make this in another green thread
            # expicially, in parallel with primary GET
            sub_req = make_subrequest(
                self.request.environ,
                'GET', resource,
                agent=self.agent)
            sub_resp = sub_req.get_response(self.app)
            sreq.extra_data_list.append(
                StorletData(
                    self._get_user_metadata(sub_resp.headers),
                    sub_resp.app_iter))
        return sreq

    @public
    def GET(self):
        """
        GET handler on Proxy
        """
        if self.is_range_request:
            msg = 'Storlet execution with range header is not supported'
            raise HTTPBadRequest(msg.encode('utf8'),
                                 request=self.request)

        params = self.verify_access_to_storlet()
        self.augment_storlet_request(params)

        # Range requests:
        # Range requests are not allowed with storlet invocation.
        # To run a storlet on a selected input range use the X-Storlet-Range
        # header.
        # If the range request is to be executed on the proxy we
        # create an HTTP Range request based on X-Storlet-Range
        # and let the request continue so that we get the required
        # range as input to the storlet that would get executed on
        # the proxy.
        if self.execute_range_on_proxy:
            self.request.headers['Range'] = \
                self.request.headers['X-Storlet-Range']

        original_resp = self.request.get_response(self.app)
        if original_resp.status_int == 403:
            # The user is unauthoried to read from the container.
            # It might be, however, that the user is permitted
            # to read given that the required storlet is executed.
            if not self.request.environ['HTTP_X_USER_NAME']:
                # The requester is not even an authenticated user.
                self.logger.info(('Storlet run request by an'
                                  ' authenticated user'))
                raise HTTPUnauthorized(b'User is not authorized')

            user_name = self.request.environ['HTTP_X_USER_NAME']
            storlet_name = self.request.headers['X-Run-Storlet']
            internal_referer = '//%s' % self._build_acl_string(user_name,
                                                               storlet_name)
            self.logger.info(('Got 403 for original GET %s request. '
                              'Trying with storlet internal referer %s' %
                              (self.path, internal_referer)))
            self.request.referrer = self.request.referer = internal_referer
            original_resp = self.request.get_response(self.app)

        if original_resp.is_success:
            # The get request may be a SLO object GET request.
            # Simplest solution would be to invoke a HEAD
            # for every GET request to test if we are in SLO case.
            # In order to save the HEAD overhead we implemented
            # a slightly more involved flow:
            # At proxy side, we augment request with Storlet stuff
            # and let the request flow.
            # At object side, we invoke the plain (non Storlet)
            # request and test if we are in SLO case.
            # and invoke Storlet only if non SLO case.
            # Back at proxy side, we test if test received
            # full object to detect if we are in SLO case,
            # and invoke Storlet only if in SLO case.
            if self.is_proxy_runnable(original_resp):
                return self.apply_storlet(original_resp)
            else:
                # Non proxy GET case: Storlet was already invoked at
                # object side
                # TODO(kota_): Do we need to pop the Transfer-Encoding/
                #              Content-Length header from the resp?
                if 'Transfer-Encoding' in original_resp.headers:
                    original_resp.headers.pop('Transfer-Encoding')

                original_resp.headers['Content-Length'] = None
                return original_resp

        else:
            # In failure case, we need nothing to do, just return original
            # response
            return original_resp

    def _validate_copy_request(self):
        # We currently block copy from account
        unsupported_headers = ['X-Copy-From-Account',
                               'Destination-Account',
                               'X-Fresh-Metadata']

        for header in unsupported_headers:
            if header in self.request.headers:
                msg = 'Storlet on copy with %s is not supported' % header
                raise HTTPBadRequest(msg.encode('utf8'))

    def handle_put_copy_response(self, app_iter):
        self._remove_storlet_headers(self.request.headers)
        if 'CONTENT_LENGTH' in self.request.environ:
            self.request.environ.pop('CONTENT_LENGTH')
        self.request.headers['Transfer-Encoding'] = 'chunked'

        self.request.environ['wsgi.input'] = FileLikeIter(app_iter)
        return self.request.get_response(self.app)

    def _remove_storlet_headers(self, headers):
        for key in list(headers):
            if (key.startswith('X-Storlet-') or
                    key.startswith('X-Object-Meta-Storlet-') or
                    key == 'X-Run-Storlet'):
                headers.pop(key)

    def base_handle_copy_request(self, src_container, src_obj,
                                 dest_container, dest_object):
        """
        Unified path for:
        PUT verb with X-Copy-From and
        COPY verb with Destination
        """
        # Get an iterator over the source object
        source_path = '/%s/%s/%s/%s' % (self.api_version, self.account,
                                        src_container, src_obj)
        source_req = self.request.copy_get()
        source_req.headers.pop('X-Backend-Storage-Policy-Index', None)
        source_req.path_info = source_path

        # In the copy case we can run either in the proxy
        # or in the object node:
        # e.g. if the object is SLO or we have extra resources
        # we run on the proxy. Otherwise, we run on the object
        # Handling in proxy means that:
        # 0. Handle in the proxy
        # 1. The object GET request to the object node
        #    should be called without 'X-Run-Storlet'
        # 2. The metadata in the response from the object node
        #    should not be prefixed with X-Object-Meta
        if self.is_proxy_runnable():
            source_req.headers.pop('X-Run-Storlet', None)

        src_resp = source_req.get_response(self.app)
        copy_headers(src_resp.headers, self.request.headers)

        # We check here again, because src_resp may reveal that
        # the object is an SLO and so even if the above check was
        # False, we now may need to run on proxy
        if self.is_proxy_runnable(src_resp):
            # We need to run on proxy.
            # Do it and fixup the user metadata headers.
            sreq = self._build_storlet_request(self.request, src_resp.headers,
                                               src_resp.app_iter)
            sresp = self.gateway.invocation_flow(sreq)
            data_iter = sresp.data.data_iter
            self._set_metadata_in_headers(self.request.headers,
                                          sresp.data.user_metadata)
        else:
            data_iter = src_resp.app_iter

        resp = self.handle_put_copy_response(data_iter)

        acct, path = src_resp.environ['PATH_INFO'].split('/', 3)[2:4]
        resp.headers['X-Storlet-Generated-From-Account'] = quote(acct)
        resp.headers['X-Storlet-Generated-From'] = quote(path)
        if 'last-modified' in src_resp.headers:
            resp.headers['X-Storlet-Generated-From-Last-Modified'] = \
                src_resp.headers['last-modified']
        return resp

    @public
    def PUT(self):
        """
        PUT handler on Proxy
        """

        params = self.verify_access_to_storlet()
        self.augment_storlet_request(params)
        if self.is_put_copy_request:
            self._validate_copy_request()
            src_container, src_obj = check_copy_from_header(self.request)
            dest_container = self.container
            dest_object = self.obj
            self.request.headers.pop('X-Copy-From', None)
            return self.base_handle_copy_request(src_container, src_obj,
                                                 dest_container, dest_object)

        # TODO(takashi): chunk size should be configurable
        reader = self.request.environ['wsgi.input'].read
        body_iter = iter(lambda: reader(65536), b'')
        sreq = self._build_storlet_request(
            self.request, self.request.headers, body_iter)

        sresp = self.gateway.invocation_flow(sreq)
        self._set_metadata_in_headers(self.request.headers,
                                      sresp.data.user_metadata)
        return self.handle_put_copy_response(sresp.data.data_iter)

    @public
    def COPY(self):
        """
        COPY handler on Proxy
        """
        if not self.request.headers.get('Destination'):
            return HTTPPreconditionFailed(request=self.request,
                                          body='Destination header required')

        params = self.verify_access_to_storlet()
        self.augment_storlet_request(params)
        self._validate_copy_request()
        dest_container, dest_object = check_destination_header(self.request)

        # re-write the existing request as a PUT instead of creating a new one
        # TODO(eranr): do we want a new sub_request or re-write existing one as
        # we do below. See proxy obj controller COPY.
        self.request.method = 'PUT'
        self.request.path_info = '/v1/%s/%s/%s' % \
                                 (self.account, dest_container, dest_object)
        self.request.headers['Content-Length'] = 0
        del self.request.headers['Destination']

        return self.base_handle_copy_request(self.container, self.obj,
                                             dest_container, dest_object)

    @public
    def POST(self):
        """
        POST handler on Proxy

        Deals with storlet ACL updates
        """
        # Get the current container's ACL
        # We perform a sub request rather than get_container_info
        # since get_container_info bypasses authorization, and we
        # prefer to be on the safe side.
        sub_req = make_subrequest(self.request.environ,
                                  'HEAD', self.path,
                                  agent=self.agent)
        sub_resp = sub_req.get_response(self.app)
        if sub_resp.status_int != 204:
            self.logger.info("Failed to retrieve container metadata")
            return HTTPUnauthorized(('Unauthorized to get or modify '
                                     'the container ACL'))

        # Add the requested ACL
        read_acl = sub_resp.headers.get("X-Container-Read")
        if read_acl:
            new_read_acl = ','.join([read_acl, self.acl_string])
        else:
            new_read_acl = self.acl_string

        self.request.headers['X-Container-Read'] = new_read_acl
        resp = self.request.get_response(self.app)
        return resp



# Copyright IBM Corp. 2015, 2015 All Rights Reserved
# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from swiftclient import client as c
from tests.functional.java import StorletJavaFunctionalTest
import unittest
from urllib.request import Request, urlopen


class TestThumbnailStorlet(StorletJavaFunctionalTest):
    def setUp(self):
        self.storlet_log = None
        self.additional_headers = {}
        main_class = 'org.openstack.storlet.thumbnail.ThumbnailStorlet'
        super(TestThumbnailStorlet, self).setUp('ThumbnailStorlet',
                                                'thumbnail-1.0.jar',
                                                main_class,
                                                'sample.jpg')

    def invoke_storlet_on_get(self):
        headers = {'X-Run-Storlet': self.storlet_name}
        headers.update(self.additional_headers)
        resp = dict()
        resp_headers, gf = c.get_object(self.url, self.token,
                                        self.container,
                                        self.storlet_file,
                                        response_dict=resp,
                                        headers=headers)
        with open('/tmp/sample.jpg', 'wb') as f:
            f.write(gf)

        self.assertIn(resp['status'], [200, 202])

    def invoke_storlet_on_put(self):
        headers = {'X-Run-Storlet': self.storlet_name,
                   'x-object-meta-name': 'thumbnail'}
        headers.update(self.additional_headers)
        resp = dict()
        source_file = '%s/%s' % (self.path_to_bundle, self.storlet_file)
        with open(source_file, 'rb') as f:
            c.put_object(self.url, self.token,
                         self.container, 'gen_thumb_on_put.jpg', f,
                         headers=headers,
                         response_dict=resp)

        status = resp.get('status')
        self.assertIn(status, [201, 202])

        headers = c.head_object(self.url, self.token,
                                self.container, 'gen_thumb_on_put.jpg')
        self.assertLess(int(headers['content-length']), 1087318)
        self.assertEqual('thumbnail', headers['x-object-meta-name'])

    def invoke_storlet_on_copy_from(self):
        headers = {'X-Run-Storlet': self.storlet_name,
                   'X-Object-Meta-Name': 'thumbnail',
                   'X-Copy-From': '%s/%s' %
                   (self.container, self.storlet_file)}
        headers.update(self.additional_headers)
        resp = dict()
        c.put_object(self.url, self.token,
                     self.container, 'gen_thumb_on_copy.jpg', '',
                     headers=headers,
                     response_dict=resp)

        status = resp.get('status')
        self.assertIn(status, [201, 202])
        rh = resp['headers']
        self.assertEqual(rh['x-storlet-generated-from'],
                         '%s/%s' %
                         (self.container, self.storlet_file))
        self.assertEqual(rh['x-storlet-generated-from-account'],
                         self.acct)
        self.assertIn('x-storlet-generated-from-last-modified', rh)

        headers = c.head_object(self.url, self.token,
                                self.container, 'gen_thumb_on_copy.jpg')
        self.assertLess(int(headers['content-length']), 1087318)
        self.assertEqual('thumbnail', headers['x-object-meta-name'])
        self.assertTrue('x-object-meta-x-timestamp' not in headers)
        self.assertTrue('x-timestamp' in headers)

    def invoke_storlet_on_copy_dest(self):
        # No COPY in swiftclient. Using urllib instead...
        url = '%s/%s/%s' % (self.url, self.container, self.storlet_file)
        headers = {'X-Auth-Token': self.token,
                   'X-Run-Storlet': self.storlet_name,
                   'X-Object-Meta-Name': 'thumbnail',
                   'Destination': '%s/gen_thumb_on_copy_.jpg' % self.container}
        headers.update(self.additional_headers)
        req = Request(url, headers=headers)
        req.get_method = lambda: 'COPY'
        conn = urlopen(req, timeout=10)
        status = conn.getcode()
        self.assertIn(status, [201, 202])

        headers = c.head_object(self.url, self.token,
                                self.container, 'gen_thumb_on_copy_.jpg')
        self.assertLess(int(headers['content-length']), 1087318)
        self.assertEqual('thumbnail', headers['x-object-meta-name'])
        self.assertTrue('x-object-meta-x-timestamp' not in headers)
        self.assertTrue('x-timestamp' in headers)

    def test_get(self):
        self.invoke_storlet_on_get()

    def test_put(self):
        self.invoke_storlet_on_put()

    def test_copy_put(self):
        self.invoke_storlet_on_copy_from()

    def test_copy(self):
        self.invoke_storlet_on_copy_dest()


class TestThumbnailStorletOnProxy(TestThumbnailStorlet):
    def setUp(self):
        super(TestThumbnailStorletOnProxy, self).setUp()
        self.additional_headers = {'X-Storlet-Run-On-Proxy': ''}


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2016 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pexpect
import swiftclient
import os
import hashlib
from storlets.tools.cluster_config_parser import ClusterConfig
from storlets.tools.utils import get_auth

EXPECT = 'expect'
SEND_LINE = 'sendline'


class DeployTestMixin(object):
    def assertUploadedFile(self, container, file_path):
        # Make sure the existence in the swift inside
        conf = ClusterConfig(self.conf_file)
        url, token = get_auth(conf, conf.admin_user, conf.admin_password)

        expected_file = os.path.basename(file_path)
        resp_headers = swiftclient.client.head_object(
            url, token, container, expected_file)
        hasher = hashlib.md5()
        with open(file_path, 'rb') as f:
            hasher.update(f.read())
        self.assertEqual(resp_headers['etag'], hasher.hexdigest())

    def test_deploy_storlet(self):
        child = pexpect.spawn('python %s %s' % (self.deploy_storlet_path,
                                                self.conf_file))
        try:
            for message, command in self.scenario:
                # FIXME(kota_): i don't get yet why the *expect* call stack
                #               stdout from print function but it seems safe
                #               to assert the output by hand rather than using
                #               expect method
                if command == EXPECT:
                    line = child.readline()
                    self.assertEqual(message.strip(), line.strip())
                elif command == SEND_LINE:
                    child.sendline(message)
                    line = child.readline()
                    self.assertEqual(message.strip(), line.strip())
                else:
                    self.fail("Unexpected scenario found %s, %s"
                              % (message, command))
        except pexpect.EOF as err:
            self.fail(
                'Expected message "%s" not found: %s' % (message, err))
        except AssertionError:
            # drain all error message lines for debug
            print(child.read())
            raise

        # Make sure the existence in the swift inside
        # for storlet app
        self.assertUploadedFile('storlet', self.execdep_storlet_file)
        # for dependency
        self.assertUploadedFile('dependency', self.execdep_storlet_dep_file)



# Copyright IBM Corp. 2015, 2015 All Rights Reserved
# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from swiftclient import client as c
from tests.functional.java import StorletJavaFunctionalTest
import unittest


class TestCompressStorlet(StorletJavaFunctionalTest):
    def setUp(self):
        self.storlet_log = ''
        self.additional_headers = {}
        main_class = 'org.openstack.storlet.compress.CompressStorlet'
        super(TestCompressStorlet, self).setUp('CompressStorlet',
                                               'compressstorlet-1.0.jar',
                                               main_class,
                                               'input.txt')

    def test_put(self):
        headers = {'X-Run-Storlet': self.storlet_name}
        headers.update(self.additional_headers)
        querystring = "action=compress"

        # simply set 1KB string data to compress
        data = b'A' * 1024

        response = dict()
        c.put_object(self.url, self.token, self.container, self.storlet_file,
                     data, None, None, None,
                     "application/octet-stream", headers, None, None,
                     querystring, response)

        querystring = "action=uncompress"

        original_headers, original_content = \
            c.get_object(self.url, self.token, self.container,
                         self.storlet_file,
                         response_dict=dict())

        object_length = int(original_headers['content-length'])

        self.assertLess(object_length, len(data))

        processed_headers, returned_content = \
            c.get_object(self.url, self.token, self.container,
                         self.storlet_file,
                         query_string=querystring, response_dict=dict(),
                         headers=headers, resp_chunk_size=object_length)
        processed_content = b''
        for chunk in returned_content:
            if chunk:
                processed_content += chunk

        self.assertEqual(data, processed_content)


class TestCompressStorletOnProxy(TestCompressStorlet):
    def setUp(self):
        super(TestCompressStorletOnProxy, self).setUp()
        self.additional_headers = {'X-Storlet-Run-On-Proxy': ''}


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2010-2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from swift.common.swob import Request, HeaderKeyDict


class FakeApp(object):

    def __init__(self):
        self._calls = []
        self._responses = {}

    def __call__(self, env, start_response):
        req = Request(env)

        # Get request parameters
        req_method = req.method
        req_path = req.path
        req_headers = req.headers
        req_body = req.body
        self._calls.append((req_method, req_path, req_headers, req_body))
        try:
            resp_cls, raw_headers, body = \
                self._responses[(req_method, req_path)]
            headers = HeaderKeyDict(raw_headers)
        except KeyError:
            if req_method == 'HEAD' and ('GET', req_path) in self._responses:
                resp_cls, raw_headers, _ = self._responses[('GET', req_path)]
                body = None
                headers = HeaderKeyDict(raw_headers)
            else:
                raise
        resp = resp_cls(req=req, headers=headers, body=body,
                        conditional_response=True)
        return resp(env, start_response)

    def register(self, method, path, resp_cls, headers=None, body=None):
        self._responses[(method, path)] = (resp_cls, headers, body)

    def reset_responses(self):
        self._responses = {}

    def reset_calls(self):
        self._calls = []

    def reset_all(self):
        self.reset_responses()
        self.reset_calls()

    def call_count(self, method, path):
        return len(self.get_calls(method, path))

    def get_calls(self, method=None, path=None):
        if not method or not path:
            return self._calls
        else:
            return([(m, p, h, b) for (m, p, h, b) in self._calls
                   if m == method and p == path])



# Copyright (c) 2016 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import sys
from storlets.sbus.client import SBusClient
from storlets.sbus.client.exceptions import SBusClientException

EXIT_SUCCESS = 0
EXIT_ERROR = 1


def main():
    argv = sys.argv

    # TODO(takashi): Add more detailed help message
    if len(argv) < 3:
        print('sbus <command> <pipe_path>')
        sys.exit(EXIT_ERROR)

    command = argv[1]
    pipe_path = argv[2]

    if not os.path.exists(pipe_path):
        print('ERROR: Pipe file %s does not exist' % pipe_path)
        sys.exit(EXIT_ERROR)

    client = SBusClient(pipe_path)
    try:
        handler = getattr(client, command)

        # TODO(takashi): Currently this only works for ping or halt.
        #                We need to pass more parameters like storlet_name
        #                to implement the other command types.
        resp = handler()
    except (AttributeError, NotImplementedError):
        print('ERROR: Command %s is not supported' % command)
        sys.exit(EXIT_ERROR)
    except SBusClientException as err:
        print('ERROR: Failed to send sbus command %s to %s: %s'
              % (command, pipe_path, err))
        sys.exit(EXIT_ERROR)
    except Exception as err:
        print('ERROR: Unknown error: %s' % err)
        sys.exit(EXIT_ERROR)

    print('Response: %s: %s' % (resp.status, resp.message))
    if resp.status:
        print('OK')
        sys.exit(EXIT_SUCCESS)
    else:
        print('ERROR: Got error response')
        sys.exit(EXIT_ERROR)



# Copyright (c) 2010-2017 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from swiftclient import client
from tests.functional.python import StorletPythonFunctionalTest
import unittest


class TestSymlink(StorletPythonFunctionalTest):
    def setUp(self):
        self.storlet_log = 'simple-symlink.log'
        self.additional_headers = {}
        self.content = b'abcdefghijklmonp'
        super(TestSymlink, self).setUp(
            storlet_dir='simple',
            storlet_name='simple.py',
            storlet_main='simple.SimpleStorlet',
            storlet_file='source.txt',
            headers={})

        symlink_target = '/'.join([self.container, self.storlet_file])
        client.put_object(self.url, self.token, self.container, 'test_link',
                          '', headers={'X-Symlink-Target': symlink_target})

    def test_get(self):
        req_headers = {'X-Run-Storlet': self.storlet_name}
        headers, content = client.get_object(
            self.url, self.token, self.container, self.storlet_file,
            headers=req_headers)
        self.assertEqual('simple', headers['x-object-meta-test'])
        self.assertEqual(self.content, content)

        req_headers = {'X-Run-Storlet': self.storlet_name}
        headers, content = client.get_object(
            self.url, self.token, self.container, 'test_link',
            headers=req_headers)
        self.assertEqual('simple', headers['x-object-meta-test'])
        self.assertEqual(self.content, content)


class TestSymlinkOnProxy(TestSymlink):
    def setUp(self):
        super(TestSymlinkOnProxy, self).setUp()
        self.additional_headers = {'X-Storlet-Run-On-Proxy': ''}


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2010-2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import os
import select
from storlets.gateway.common.exceptions import StorletTimeout


class FileDescriptorIterator(object):
    def __init__(self, fd, timeout, cancel_func):
        self.closed = False
        self.fd = fd
        self.timeout = timeout
        self.cancel_func = cancel_func
        self.buf = b''

    def __iter__(self):
        return self

    def read_with_timeout(self, size):
        try:
            with StorletTimeout(self.timeout):
                chunk = os.read(self.fd, size)
        except StorletTimeout:
            if self.cancel_func:
                self.cancel_func()
            self.close()
            raise
        except Exception:
            self.close()
            raise
        return chunk

    def next(self, size=64 * 1024):
        if len(self.buf) < size:
            r, w, e = select.select([self.fd], [], [], self.timeout)
            if len(r) == 0:
                self.close()

            if self.fd in r:
                self.buf += self.read_with_timeout(size - len(self.buf))
                if self.buf == b'':
                    raise StopIteration('Stopped iterator ex')
            else:
                raise StopIteration('Stopped iterator ex')

        if len(self.buf) > size:
            data = self.buf[:size]
            self.buf = self.buf[size:]
        else:
            data = self.buf
            self.buf = b''
        return data

    __next__ = next

    def _close_check(self):
        if self.closed:
            raise ValueError('I/O operation on closed file')

    def read(self, size=64 * 1024):
        self._close_check()
        return self.next(size)

    def readline(self, size=-1):
        self._close_check()

        # read data into self.buf if there is not enough data
        while b'\n' not in self.buf and \
              (size < 0 or len(self.buf) < size):
            if size < 0:
                chunk = self.read()
            else:
                chunk = self.read(size - len(self.buf))
            if not chunk:
                break
            self.buf += chunk

        # Retrieve one line from buf
        data, sep, rest = self.buf.partition(b'\n')
        data += sep
        self.buf = rest

        # cut out size from retrieved line
        if size >= 0 and len(data) > size:
            self.buf = data[size:] + self.buf
            data = data[:size]

        return data

    def readlines(self, sizehint=-1):
        self._close_check()
        lines = []
        try:
            while True:
                line = self.readline(sizehint)
                if not line:
                    break
                lines.append(line)
                if sizehint >= 0:
                    sizehint -= len(line)
                    if sizehint <= 0:
                        break
        except StopIteration:
            pass
        return lines

    def close(self):
        if self.closed:
            return
        os.close(self.fd)
        self.closed = True

    def __del__(self):
        self.close()


class StorletData(object):
    def __init__(self, user_metadata, data_iter=None, data_fd=None,
                 timeout=10, cancel=None):
        if data_iter is None and data_fd is None:
            raise ValueError('Either of data_iter or data_fd should not be '
                             'None')
        if cancel is not None and data_iter is not None:
            raise ValueError('cancel func can only be specified with data_fd')
        self.user_metadata = user_metadata
        self.data_fd = data_fd
        self._data_iter = data_iter
        self.timeout = timeout
        self.cancel = cancel

    @property
    def data_iter(self):
        if self._data_iter is None:
            self._data_iter = FileDescriptorIterator(
                self.data_fd, self.timeout, self.cancel)
        return self._data_iter

    @property
    def has_fd(self):
        return self.data_fd is not None


class StorletRequest(object):
    """
    The StorletRequest class represents a request to be processed by
    the storlet.
    """

    required_options = []

    def __init__(self, storlet_id, params, data, options=None,
                 extra_data_list=None):
        """
        :param storlet_id: storlet id
        :param params: parameters for storlet execution
        :param data: StorletData instance
        :param options: options specific to StorletRequest types
        :param extra_data_list: List of StorletData instances
        :raises ValueError: when some of the required options are missing
        """
        self.storlet_id = storlet_id
        self.params = copy.deepcopy(params)
        self.data = data
        self.options = options or {}
        self.extra_data_list = extra_data_list or []

        for opt in self.required_options:
            if options.get(opt) is None:
                raise ValueError('Required option %s is missing' % opt)


class StorletResponse(object):
    """
    The StorletResponse class represents a response from the storlet
    """

    def __init__(self, data):
        self.data = data



# Copyright IBM Corp. 2015, 2015 All Rights Reserved
# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import random
import string
from swiftclient import client
from tests.functional.java import StorletJavaFunctionalTest
import unittest


class TestSLO(StorletJavaFunctionalTest):
    def setUp(self):
        self.storlet_log = 'identitystorlet-1.0.log'
        self.additional_headers = {}
        self.chunks = []
        main_class = 'org.openstack.storlet.identity.IdentityStorlet'
        super(TestSLO, self).setUp('IdentityStorlet',
                                   'identitystorlet-1.0.jar',
                                   main_class,
                                   '')

        self.create_local_chunks()
        self.put_SLO()

    def create_local_chunks(self):
        for i in range(10):
            self.chunks.append(
                ''.join([random.choice(string.ascii_uppercase + string.digits)
                         for _ in range(1024)]).encode('ascii'))

    def put_SLO(self):
        assembly = []
        for i in range(10):
            oname = 'slo_chunk_%d' % i
            etag = client.put_object(self.url, self.token,
                                     self.container, oname, self.chunks[i],
                                     content_type="application/octet-stream")

            segment = dict()
            segment['path'] = '%s/%s' % (self.container, oname)
            segment['size_bytes'] = 1024
            segment['etag'] = etag
            assembly.append(segment)

        headers = {'x-object-meta-prop1': 'val1'}
        client.put_object(self.url, self.token, self.container,
                          'assembly', json.dumps(assembly),
                          headers=headers,
                          query_string='multipart-manifest=put')

    def compare_slo_to_chunks(self, body):
        length = 0
        for (i, chunk) in enumerate(body):
            self.assertEqual(chunk, self.chunks[i])
            length += 1
        self.assertEqual(length, 10)

    def test_get_SLO_without_storlet(self):
        _, body = client.get_object(self.url, self.token,
                                    self.container, 'assembly',
                                    resp_chunk_size=1024)
        self.compare_slo_to_chunks(body)

    test_get_SLO_without_storlet.slow = 1

    def test_get_SLO_with_storlet(self):
        headers = {'X-Run-Storlet': self.storlet_name}
        headers.update(self.additional_headers)
        _, body = client.get_object(self.url, self.token,
                                    self.container, 'assembly',
                                    resp_chunk_size=1024,
                                    headers=headers)
        self.compare_slo_to_chunks(body)

    test_get_SLO_with_storlet.slow = 1


class TestSloOnProxy(TestSLO):
    def setUp(self):
        super(TestSloOnProxy, self).setUp()
        self.additional_headers = {'X-Storlet-Run-On-Proxy': ''}


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from swiftclient import client as c
from tests.functional.java import StorletJavaFunctionalTest
import unittest
from hashlib import md5


class TestMultiInputStorlet(StorletJavaFunctionalTest):
    def setUp(self):
        self.additional_headers = {}
        super(TestMultiInputStorlet, self).setUp(
            storlet_dir='MultiInputStorlet',
            storlet_name='multiinputstorlet-1.0.jar',
            storlet_main='org.openstack.storlet.multiinput.MultiInputStorlet',
            storlet_file=None,
            headers={})

    def test_get_extra_sources(self):
        obj = 'small'
        obj2 = 'small2'
        c.put_object(self.url, self.token,
                     self.container, obj,
                     b'0123456789abcd')
        c.put_object(self.url, self.token,
                     self.container, obj2,
                     b'efghijklmnopqr')

        headers = {
            'X-Run-Storlet': self.storlet_name,
            'X-Storlet-Extra-Resources':
            '/%s/%s' % (self.container, obj2),
            'X-Storlet-Run-On-Proxy': ''
        }
        headers.update(self.additional_headers)

        resp_headers, resp_content = c.get_object(
            self.url, self.token, self.container, obj,
            headers=headers)
        self.assertEqual(b'0123456789abcdefghijklmnopqr',
                         resp_content)

    def test_put_extra_sources(self):
        obj = 'small'
        obj2 = 'small2'
        c.put_object(self.url, self.token,
                     self.container, obj2,
                     b'efghijklmnopqr')

        headers = {
            'X-Run-Storlet': self.storlet_name,
            'X-Storlet-Extra-Resources':
            '/%s/%s' % (self.container, obj2),
            'X-Storlet-Run-On-Proxy': ''
        }
        headers.update(self.additional_headers)
        c.put_object(
            self.url, self.token, self.container, obj,
            b'0123456789abcd', headers=headers)

        resp_headers, resp_content = c.get_object(
            self.url, self.token, self.container, obj)
        self.assertEqual(b'0123456789abcdefghijklmnopqr',
                         resp_content)

    def test_put_x_copy_from_extra_sources(self):
        obj = 'small'
        obj2 = 'small2'
        copied_obj = 'copied'
        c.put_object(self.url, self.token,
                     self.container, obj,
                     b'0123456789abcd',
                     headers={'X-Object-Meta-Key1': 'value1'})
        c.put_object(self.url, self.token,
                     self.container, obj2,
                     b'efghijklmnopqr',
                     headers={'X-Object-Meta-Key2': 'value2'})

        headers = {
            'X-Run-Storlet': self.storlet_name,
            'X-Copy-From':
            '/%s/%s' % (self.container, obj),
            'X-Storlet-Extra-Resources':
            '/%s/%s' % (self.container, obj2),
        }
        headers.update(self.additional_headers)

        expected_string = b'0123456789abcdefghijklmnopqr'
        etag = c.put_object(
            self.url, self.token, self.container, copied_obj,
            headers=headers)

        hasher = md5()
        hasher.update(expected_string)

        self.assertEqual(hasher.hexdigest(), etag)

        resp_headers, resp_content = c.get_object(
            self.url, self.token, self.container, copied_obj)
        self.assertEqual(expected_string, resp_content)
        self.assertEqual('value1', resp_headers['x-object-meta-key1'])
        self.assertEqual('value2', resp_headers['x-object-meta-key2'])


class TestMultiInputStorletOnProxy(TestMultiInputStorlet):
    def setUp(self):
        super(TestMultiInputStorletOnProxy, self).setUp()
        self.additional_headers = {'X-Storlet-Run-On-Proxy': ''}


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import string
from swiftclient import client
from tests.functional.python import StorletPythonFunctionalTest
import unittest


class TestExecQueryHeaderStorlet(StorletPythonFunctionalTest):
    def setUp(self):
        self.additional_headers = {}
        super(TestExecQueryHeaderStorlet, self).setUp(
            storlet_dir='exec_query_header',
            storlet_name='exec_query_header.py',
            storlet_main='exec_query_header.ExecQueryHeaderStorlet',
            storlet_file='source.txt',
            headers={})

    def invoke_storlet(self, op, params=None, header_parameters=False):
        headers = {'X-Run-Storlet': self.storlet_name}
        headers.update(self.additional_headers)
        if params is not None:
            if header_parameters:
                querystring = None
                count = 1
                for key in params:
                    headers['X-Storlet-Parameter-' + str(count)] = \
                        key + ':' + params[key]
                    count = count + 1
            else:
                querystring = '&'.join(['%s=%s' % (k, v)
                                       for k, v in params.items()])
        else:
            querystring = None

        if op == 'GET':
            # Get original object
            original_h, original_c = client.get_object(self.url, self.token,
                                                       self.container,
                                                       self.storlet_file,
                                                       response_dict=dict())
            file_length = int(original_h['content-length'])
            processed_h, returned_c = client.get_object(
                self.url, self.token,
                self.container, self.storlet_file,
                query_string=querystring, response_dict=dict(),
                headers=headers, resp_chunk_size=file_length)
            processed_c = b''
            for chunk in returned_c:
                if chunk:
                    processed_c += chunk

            for param_key, param_value in params.items():
                mdv = processed_h['X-Object-Meta-%s'.lower() % param_key]
                self.assertEqual(mdv, param_value)

        if op == 'PUT':
            # PUT a random file
            response = dict()
            uploaded_c = ''.join(
                random.choice(string.ascii_uppercase + string.digits)
                for _ in range(1024)).encode('ascii')
            content_length = None
            client.put_object(self.url, self.token, self.container,
                              'random_source',
                              uploaded_c, content_length, None, None,
                              "application/octet-stream", headers, None, None,
                              querystring, response)
            resp_headers, saved_c = client.get_object(self.url, self.token,
                                                      self.container,
                                                      'random_source',
                                                      response_dict=dict())

            for param_key, param_value in params.items():
                mdv = resp_headers['X-Object-Meta-%s'.lower() % param_key]
                self.assertEqual(mdv, param_value)

    def test_put_execute_with_header(self):
        self.invoke_storlet('PUT', {'key1': 'value1'},
                            header_parameters=True)
        self.invoke_storlet('PUT', {'key1': 'value1', 'key2': 'value2'},
                            header_parameters=True)

    def test_put_execute_with_query(self):
        self.invoke_storlet('PUT', {'key1': 'value1'})
        self.invoke_storlet('PUT', {'key1': 'value1', 'key2': 'value2'})

    def test_get_execute_with_header(self):
        self.invoke_storlet('GET', {'key1': 'value1'},
                            header_parameters=True)
        self.invoke_storlet('GET', {'key1': 'value1', 'key2': 'value2'},
                            header_parameters=True)

    def test_get_execute_with_query(self):
        self.invoke_storlet('GET', {'key1': 'value1'})
        self.invoke_storlet('GET', {'key1': 'value1', 'key2': 'value2'})


class TestExecQueryHeaderStorletOnProxy(TestExecQueryHeaderStorlet):
    def setUp(self):
        super(TestExecQueryHeaderStorletOnProxy, self).setUp()
        self.additional_headers = {'X-Storlet-Run-On-Proxy': ''}


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2015-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest
from unittest import mock

from storlets.sbus import command as sbus_cmd
from storlets.agent.daemon.server import StorletDaemon, StorletDaemonLoadError

from tests.unit import FakeLogger
from tests.unit.agent.common import test_server


class FakeModule(object):
    FakeClass = mock.MagicMock


class TestStorletDaemon(unittest.TestCase):
    def setUp(self):
        self.logger = FakeLogger()

    def test_invalid_module_name(self):
        for module_name in ('only_module',
                            'OnlyClass',
                            'nested.module.Class',
                            1, None, True, False):
            self._test_invalid_module_name(module_name)

    def _test_invalid_module_name(self, module_name):
        with self.assertRaises(ValueError) as cm:
            StorletDaemon(module_name, 'fake_path', self.logger, 16)
        self.assertEqual('Invalid storlet name %s' % module_name,
                         cm.exception.args[0])

    def test_module_not_found(self):
        with self.assertRaises(StorletDaemonLoadError) as cm:
            StorletDaemon('nomodule.Nothing', 'fake_path', self.logger, 16)
        self.assertEqual('Failed to load storlet nomodule.Nothing',
                         cm.exception.args[0])


class TestStorletDaemonMain(test_server.TestSBusServerMain):

    def _get_test_server(self):
        with mock.patch('importlib.import_module') as fake_import:
            fake_import.return_value = FakeModule()
            server = StorletDaemon(
                'fakeModule.FakeClass', self.sbus_path, self.logger, 16)
        return server

    def test_main_loop_successful_stop(self):
        # SBUS_CMD_HALT is for working to stop requested from
        # storlet_middleware
        self._test_main_loop_stop(sbus_cmd.SBUS_CMD_HALT)

    def test_main_loop_canceled_stop(self):
        # SBUS_CMD_CANCEL is for working to stop from sort of daemon
        # management tools
        # TODO(kota_): SBUS_CMD_CANCEL has more tasks to do for cleanup
        # so need more assertions.
        self._test_main_loop_stop(sbus_cmd.SBUS_CMD_CANCEL)

    def test_main_loop_timeout(self):
        self._test_main_loop_timeout()


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2015-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from logging.handlers import SysLogHandler


def get_logger(logger_name, log_level, container_id):
    """
    Initialize logging of this process and set logger format

    :param logger_name: The name to report with
    :param log_level: The verbosity level. This should be selected
    :param container_id: container id
    """
    log_level = log_level.upper()

    # NOTE(takashi): currently logging.WARNING is defined as the same value
    #                as logging.WARN, so we can properly handle WARNING here
    try:
        level = getattr(logging, log_level)
    except AttributeError:
        level = logging.ERROR

    logger = logging.getLogger("CONT #" + container_id + ": " + logger_name)

    if log_level == 'OFF':
        logging.disable(logging.CRITICAL)
    else:
        logger.setLevel(level)

    log_handler = SysLogHandler('/dev/log')
    str_format = '%(name)-12s: %(levelname)-8s %(funcName)s' + \
                 ' %(lineno)s [%(process)d, %(threadName)s]' + \
                 ' %(message)s'
    formatter = logging.Formatter(str_format)
    log_handler.setFormatter(formatter)
    log_handler.setLevel(level)
    logger.addHandler(log_handler)
    return logger



# Copyright IBM Corp. 2015, 2015 All Rights Reserved
# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import random
import string
from swiftclient import client as c
from contextlib import contextmanager
from tests.functional.java import StorletJavaFunctionalTest
import unittest


class TestIdentityStorlet(StorletJavaFunctionalTest):
    def setUp(self):
        self.storlet_log = 'identitystorlet-1.0.log'
        self.additional_headers = {}
        main_class = 'org.openstack.storlet.identity.IdentityStorlet'
        headers = {'X-Object-Meta-Testkey': 'tester'}
        super(TestIdentityStorlet, self).setUp('IdentityStorlet',
                                               'identitystorlet-1.0.jar',
                                               main_class,
                                               'source.txt',
                                               ['get42'],
                                               headers)

    def invoke_storlet(self, op, params=None, global_params=None,
                       header_parameters=False):
        headers = {'X-Run-Storlet': self.storlet_name}
        headers.update(self.additional_headers)
        if params is not None:
            if header_parameters:
                querystring = None
                count = 1
                for key in params:
                    headers['X-Storlet-Parameter-' + str(count)] = \
                        key + ':' + params[key]
                    count = count + 1
            else:
                querystring = '&'.join(['%s=%s' % (k, v)
                                       for k, v in params.items()])
        else:
            querystring = None

        if op == 'GET':
            # Get original object
            original_h, original_c = c.get_object(self.url, self.token,
                                                  self.container,
                                                  self.storlet_file,
                                                  response_dict=dict())
            # print original_headers
            file_length = int(original_h['content-length'])
            processed_h, returned_c = c.get_object(
                self.url, self.token,
                self.container, self.storlet_file,
                query_string=querystring, response_dict=dict(),
                headers=headers, resp_chunk_size=file_length)
            processed_c = b''
            for chunk in returned_c:
                if chunk:
                    processed_c += chunk

            if params is not None and params.get('execute', None) is not None:
                mdv = processed_h['X-Object-Meta-Execution-Result'.lower()]
                self.assertEqual('42', mdv)
            if params is not None and params.get('double', None) == 'true':
                self.assertEqual(original_c, processed_c[:file_length])
                self.assertEqual(original_c, processed_c[file_length:])
            else:
                self.assertEqual(original_c, processed_c)
            self.assertEqual(original_h['X-Object-Meta-Testkey'.lower()],
                             processed_h['X-Object-Meta-Testkey'.lower()])

        if op == 'PUT':
            # PUT a random file
            response = dict()
            uploaded_c = ''.join(
                random.choice(string.ascii_uppercase + string.digits)
                for _ in range(1024)).encode('ascii')
            random_md = ''.join(random.choice(string.ascii_uppercase +
                                string.digits) for _ in range(32))
            content_length = None
            headers['X-Object-Meta-Testkey'] = random_md
            c.put_object(self.url, self.token, self.container,
                         'identity_random_source',
                         uploaded_c, content_length, None, None,
                         "application/octet-stream", headers, None, None,
                         querystring, response)
            resp_headers, saved_c = c.get_object(self.url, self.token,
                                                 self.container,
                                                 'identity_random_source',
                                                 response_dict=dict())

            if params is not None and params.get('double', None) == 'true':
                assert(uploaded_c == saved_c[:1024])
                assert(uploaded_c == saved_c[1024:])
            else:
                assert(uploaded_c == saved_c)

            if params is not None and params.get('execute', None) is not None:
                mdv = resp_headers['X-Object-Meta-Execution-Result'.lower()]
                self.assertEqual('42', mdv)

            self.assertEqual(random_md,
                             resp_headers['X-Object-Meta-Testkey'.lower()])

    @contextmanager
    def _filecontext(self, path):
        try:
            yield
        finally:
            try:
                os.remove(path)
            except Exception:
                pass

    def test_put_10MB_file(self):
        gf_file_path = '/tmp/10MB_file'
        with self._filecontext(gf_file_path):
            TenMBFile = open('/tmp/10MB_file', 'wb')
            for _ in range(10):
                uploaded_content = ''.join('1' for _ in range(1 * 1024 * 1024))
                TenMBFile.write(uploaded_content.encode("utf-8"))
            TenMBFile.close()

            headers = {'X-Run-Storlet': self.storlet_name}
            headers.update(self.additional_headers)
            TenMBFile = open('/tmp/10MB_file', 'rb')
            response = dict()
            c.put_object(self.url, self.token,
                         self.container, '10MBFile', TenMBFile,
                         10 * 1024 * 1024, None, None,
                         "application/octet-stream",
                         headers, None, None, None, response)
            status = response.get('status')
            self.assertIn(status, [200, 201])

    test_put_10MB_file.slow = 1

    def test_put(self):
        self.invoke_storlet('PUT')

    def test_put_execute(self):
        self.invoke_storlet('PUT', {'execute': 'true'})
        self.invoke_storlet('PUT', {'execute': 'true'},
                            header_parameters=True)

    def test_put_double(self):
        self.invoke_storlet('PUT', {'double': 'true'})
        self.invoke_storlet('PUT', {'double': 'true'},
                            header_parameters=True)

    def test_get(self):
        self.invoke_storlet('GET')

    def test_get_double(self):
        self.invoke_storlet('GET', {'double': 'true'})
        self.invoke_storlet('GET', {'double': 'true'},
                            header_parameters=True)

    def test_get_execute(self):
        self.invoke_storlet('GET', {'execute': 'true'})
        self.invoke_storlet('GET', {'execute': 'true'},
                            header_parameters=True)

    def _test_get_range(self, start, end, expected):
        srange = 'bytes=%d-%d' % (start, end)
        headers = {'X-Run-Storlet': self.storlet_name,
                   'X-Storlet-Range': srange}
        headers.update(self.additional_headers)
        junk, content = c.get_object(self.url, self.token,
                                     self.container,
                                     'small',
                                     headers=headers,
                                     response_dict=dict())
        self.assertEqual(expected, content)

    def test_get_ranges(self):
        response = dict()
        c.put_object(self.url, self.token,
                     self.container, 'small',
                     b'0123456789abcd',
                     response_dict=response)

        self._test_get_range(0, 0, b'0')
        self._test_get_range(0, 10, b'0123456789a')
        self._test_get_range(2, 10, b'23456789a')
        self._test_get_range(10, 13, b'abcd')
        self._test_get_range(10, 14, b'abcd')
        self._test_get_range(10, 15, b'abcd')


class TestIdentityStorletOnProxy(TestIdentityStorlet):
    def setUp(self):
        super(TestIdentityStorletOnProxy, self).setUp()
        self.additional_headers = {'X-Storlet-Run-On-Proxy': ''}


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
from swiftclient import client as swift_client
from tests.functional import StorletBaseFunctionalTest
import unittest


class TestSetACL(StorletBaseFunctionalTest):
    def setUp(self):
        super(TestSetACL, self).setUp()
        self.container = str(uuid.uuid4())
        swift_client.put_container(self.url,
                                   self.token,
                                   self.container)

    def tearDown(self):
        swift_client.delete_container(self.url,
                                      self.token,
                                      self.container)

    def test_set_acl(self):
        headers = {'X-Container-Read': 'adam'}
        swift_client.post_container(self.url,
                                    self.token,
                                    self.container,
                                    headers)

        headers = {'X-Storlet-Container-Read': 'john',
                   'X-Storlet-Name': 'mystorlet-1.0.jar'}
        swift_client.post_container(self.url,
                                    self.token,
                                    self.container,
                                    headers)

        headers = swift_client.head_container(self.url,
                                              self.token,
                                              self.container)
        read_acl = headers['x-container-read']
        expected_acl = ('adam,.r:storlets'
                        '.john_mystorlet-1.0.jar')
        self.assertEqual(expected_acl, read_acl)


if __name__ == '__main__':
    unittest.main()



# Copyright IBM Corp. 2015, 2015 All Rights Reserved
# Copyright (c) 2010-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import configparser


class ClusterConfig(object):

    def __init__(self, config_path):
        config = configparser.ConfigParser()
        config.read(config_path)
        options = config.options('general')
        self.conf = {}
        for option in options:
            self.conf[option] = config.get('general', option)
        self._auth_version = '3'

    # TODO(eran) get rid of returning raw conf
    def get_conf(self):
        return self.conf

    @property
    def domain_name(self):
        return self.conf['keystone_default_domain']

    @property
    def auth_uri(self):
        return self.conf['keystone_public_url']

    @property
    def project_name(self):
        return self.conf['storlets_default_project_name']

    @property
    def admin_user(self):
        return self.conf['storlets_default_project_user_name']

    @property
    def admin_password(self):
        return self.conf['storlets_default_project_user_password']

    @property
    def member_user(self):
        return self.conf['storlets_default_project_member_user']

    @property
    def member_password(self):
        return self.conf['storlets_default_project_member_password']

    @property
    def region(self):
        return self.conf.get('region', '')

    # TODO(eranr) move to cluster_config
    @property
    def auth_version(self):
        return self._auth_version



# Copyright (c) 2015-2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
import tempfile
import unittest
from storlets.agent.daemon.files import StorletFile, StorletInputFile, \
    StorletRangeInputFile, StorletOutputFile


class TestStorletFile(unittest.TestCase):
    def setUp(self):
        self.fd, self.fname = tempfile.mkstemp()
        self._prepare_file()
        self.sfile = self._create_file()

    def _prepare_file(self):
        pass

    def tearDown(self):
        try:
            # Make sure that fd is closed
            os.close(self.fd)
        except OSError:
            pass
        os.unlink(self.fname)

    def _create_file(self):
        return StorletFile(self.fd)

    def test_closed(self):
        self.assertFalse(self.sfile.closed)
        self.sfile.close()
        self.assertTrue(self.sfile.closed)

    def test_contextmanager(self):
        with self.sfile as sfile:
            self.assertFalse(sfile.closed)
        self.assertTrue(sfile.closed)


class TestStorletOutputFile(TestStorletFile):

    def setUp(self):
        self.metadata = {'key1': 'value1'}
        self.md_fd, self.md_fname = tempfile.mkstemp()
        super(TestStorletOutputFile, self).setUp()

    def tearDown(self):
        try:
            # Make sure that fd is closed
            os.close(self.md_fd)
        except OSError:
            pass
        os.unlink(self.md_fname)
        super(TestStorletOutputFile, self).tearDown()

    def _create_file(self):
        return StorletOutputFile(self.md_fd, self.fd)

    def test_set_metadata(self):
        with self.sfile as sfile:
            sfile.set_metadata(self.metadata)
            self.assertEqual(self.metadata, sfile.get_metadata())

        with open(self.md_fname, 'r') as f:
            self.assertEqual(self.metadata,
                             json.loads(f.read()))

    def test_set_metadata_twice(self):
        with self.sfile as sfile:
            sfile.set_metadata(self.metadata)
            with self.assertRaises(IOError):
                sfile.set_metadata({})

    def test_get_metadata_before_set(self):
        with self.sfile as sfile:
            with self.assertRaises(IOError):
                sfile.get_metadata()

    def test_write_before_set_metadata(self):
        with self.sfile as sfile:
            with self.assertRaises(IOError):
                sfile.write('foo')

    def test_write(self):
        with self.sfile as sfile:
            sfile.set_metadata({})
            sfile.write(b'testing')
            sfile.flush()

        with open(self.fname, 'rb') as f:
            self.assertEqual(b'testing', f.read())

    def test_writelines(self):
        with self.sfile as sfile:
            sfile.set_metadata({})
            sfile.writelines([b'tes', b'ti', b'ng'])
            sfile.flush()

        with open(self.fname, 'rb') as f:
            self.assertEqual(b'testing', f.read())


class TestStorletInputFile(TestStorletFile):

    def setUp(self, content=None):
        self.content = content or b'abcd\nefg\nhi\nj'
        self.metadata = {'key1': 'value1'}
        super(TestStorletInputFile, self).setUp()

    def _prepare_file(self):
        with open(self.fname, 'wb') as f:
            f.write(self.content)

    def _create_file(self):
        return StorletInputFile(self.metadata, self.fd)

    def test_read(self):
        self.assertEqual(b'abcd\nefg\nhi\nj', self.sfile.read())

    def test_read_size(self):
        expects = [b'abc', b'd\ne', b'fg\n', b'hi\n', b'j', b'']
        for expect in expects:
            self.assertEqual(expect, self.sfile.read(3))

    def test_readline(self):
        expects = [b'abcd\n', b'efg\n', b'hi\n', b'j', b'']
        for expect in expects:
            self.assertEqual(expect, self.sfile.readline())

    def test_readline_size(self):
        expects = [b'abc', b'd\n', b'efg', b'\n', b'hi\n', b'j', b'']
        for expect in expects:
            self.assertEqual(expect, self.sfile.readline(3))

    def test_readlines(self):
        self.assertEqual([b'abcd\n', b'efg\n', b'hi\n', b'j'],
                         self.sfile.readlines())

    def test_iter(self):
        buf = b''
        for rbuf in self.sfile:
            buf += rbuf
        self.assertEqual(b'abcd\nefg\nhi\nj', buf)

    def test_get_metadata(self):
        self.assertEqual(self.metadata, self.sfile.get_metadata())

    def test_set_metadata(self):
        with self.assertRaises(NotImplementedError):
            self.sfile.set_metadata({})


class TestStorletRangeInputFile(TestStorletInputFile):

    def setUp(self):
        content = b'012\n345\nabcd\nefg\nhi\nj67\n8'
        self.start = 8
        self.end = 21
        super(TestStorletRangeInputFile, self).setUp(content)

    def _create_file(self):
        return StorletRangeInputFile(self.metadata, self.fd, self.start,
                                     self.end)


if __name__ == '__main__':
    unittest.main()
